"""
Indicator Utilities
===================

The `Indicator` class wraps indices computations with pre- and post-processing functionality. Prior to computations,
the class runs data and metadata health checks. After computations, the class masks values that should be considered
missing and adds metadata attributes to the  object.

There are many ways to construct indicators. A good place to start is
`this notebook <notebooks/extendxclim.ipynb#Defining-new-indicators>`_.

Dictionary and YAML parser
--------------------------

To construct indicators dynamically, xclim can also use dictionaries and parse them from YAML files.
This is especially useful for generating whole indicator "submodules" from files.
This functionality is inspired by the work of `clix-meta <https://github.com/clix-meta/clix-meta/>`_.

YAML file structure
~~~~~~~~~~~~~~~~~~~

Indicator-defining yaml files are structured in the following way.
Most entries of the `indicators` section are mirroring attributes of
the :py:class:`Indicator`, please refer to its documentation for more
details on each.

.. code-block:: yaml

    module: <module name>  # Defaults to the file name
    realm: <realm>  # If given here, applies to all indicators that do not already provide it.
    keywords: <keywords> # Merged with indicator-specific keywords (joined with a space)
    references: <references> # Merged with indicator-specific references (joined with a new line)
    base: <base indicator class>  # Defaults to "Daily" and applies to all indicators that do not give it.
    doc: <module docstring>  # Defaults to a minimal header, only valid if the module doesn't already exist.
    variables:  # Optional section if indicators declared below rely on variables unknown to xclim
                # (not in `xclim.core.utils.VARIABLES`)
                # The variables are not module-dependent and will overwrite any already existing with the same name.
      <varname>:
        canonical_units: <units> # required
        description: <description> # required
        standard_name: <expected standard_name> # optional
        cell_methods: <expected cell_methods> # optional
    indicators:
      <identifier>:
        # From which Indicator to inherit
        base: <base indicator class>  # Defaults to module-wide base class
                                      # If the name startswith a '.', the base class is taken from the current module
                                      # (thus an indicator declared _above_).
                                      # Available classes are listed in `xclim.core.indicator.registry` and
                                      # `xclim.core.indicator.base_registry`.

        # General metadata, usually parsed from the `compute`'s docstring when possible.
        realm: <realm>  # defaults to module-wide realm. One of "atmos", "land", "seaIce", "ocean".
        title: <title>
        abstract: <abstract>
        keywords: <keywords>  # Space-separated, merged to module-wide keywords.
        references: <references>  # newline-seperated, merged to module-wide references.
        notes: <notes>

        # Other options
        missing: <missing method name>
        missing_options:
            # missing options mapping
        allowed_periods: [<list>, <of>, <allowed>, <periods>]

        # Compute function
        compute: <function name>  # Referring to a function in `Indices` module (xclim.indices.generic or xclim.indices)
        input:  # When "compute" is a generic function, this is a mapping from argument name to the expected variable.
                # This will allow the input units and CF metadata checks to run on the inputs.
                # Can also be used to modify the expected variable, as long as it has the same dimensionality
                # Ex: tas instead of tasmin.
                # Can refer to a variable declared in the `variables` section above.
          <var name in compute> : <variable official name>
          ...
        parameters:
         <param name>: <param data>  # Simplest case, to inject parameters in the compute function.
         <param name>:  # To change parameters metadata or to declare units when "compute" is a generic function.
            units: <param units>  # Only valid if "compute" points to a generic function
            default : <param default>
            description: <param description>
            name : <param name>  # Change the name of the parameter (similar to what `input` does for variables)
            kind: <param kind> # Override the parameter kind.
                               # This is mostly useful for transforming an optional variable into a required one by passing ``kind: 0``.
        ...
      ...  # and so on.

All fields are optional. Other fields found in the yaml file will trigger errors in xclim.
In the following, the section under `<identifier>` is referred to as `data`. When creating indicators from
a dictionary, with :py:meth:`Indicator.from_dict`, the input dict must follow the same structure of `data`.

Note that kwargs-like parameters like ``indexer`` must be injected as a dictionary (``param data`` above should be a dictionary).

When a module is built from a yaml file, the yaml is first validated against the schema (see xclim/data/schema.yml)
using the YAMALE library (:cite:p:`lopker_yamale_2022`). See the "Extending xclim" notebook for more info.

Inputs
~~~~~~
As xclim has strict definitions of possible input variables (see :py:data:`xclim.core.utils.variables`),
the mapping of `data.input` simply links an argument name from the function given in "compute"
to one of those official variables.

"""

from __future__ import annotations

import re
import warnings
import weakref
from collections import OrderedDict, defaultdict
from collections.abc import Callable, Sequence
from copy import deepcopy
from dataclasses import asdict, dataclass
from functools import reduce
from inspect import Parameter as _Parameter
from inspect import Signature, signature
from inspect import _empty as _empty_default  # noqa
from os import PathLike
from pathlib import Path
from types import ModuleType
from typing import Any

import numpy as np
import xarray
import yamale
from xarray import DataArray, Dataset
from yaml import safe_load

from xclim import indices
from xclim.core import datachecks
from xclim.core._exceptions import (
    MissingVariableError,
    ValidationError,
    raise_warn_or_log,
)
from xclim.core._types import VARIABLES
from xclim.core.calendar import parse_offset, select_time
from xclim.core.cfchecks import cfcheck_from_name
from xclim.core.formatting import (
    AttrFormatter,
    default_formatter,
    gen_call_string,
    generate_indicator_docstring,
    get_percentile_metadata,
    merge_attributes,
    parse_doc,
    update_history,
)
from xclim.core.locales import (
    TRANSLATABLE_ATTRS,
    get_local_attrs,
    get_local_formatter,
    load_locale,
    read_locale_file,
)
from xclim.core.options import (
    AS_DATASET,
    CHECK_MISSING,
    KEEP_ATTRS,
    METADATA_LOCALES,
    MISSING_METHODS,
    MISSING_OPTIONS,
    OPTIONS,
)
from xclim.core.units import check_units, convert_units_to, declare_units, units
from xclim.core.utils import (
    InputKind,
    infer_kind_from_parameter,
    is_percentile_dataarray,
    load_module,
    split_auxiliary_coordinates,
)

# Indicators registry
registry = {}  # Main class registry
base_registry = {}
_indicators_registry = defaultdict(list)  # Private instance registry


# Sentinel class for unset properties of Indicator's parameters."""
class _empty:  # pylint: disable=too-few-public-methods
    pass


@dataclass
class Parameter:
    """Class for storing an indicator's controllable parameter.

    For convenience, this class implements a special "contains".

    Example
    -------
    >>> p = Parameter(InputKind.NUMBER, default=2, description="A simple number")
    >>> p.units is Parameter._empty  # has not been set
    True
    >>> "units" in p  # Easier/retrocompatible way to test if units are set
    False
    >>> p.description
    'A simple number'
    """

    _empty = _empty

    kind: InputKind
    default: Any = _empty_default
    # Name of the compute function's argument corresponding to this parameter.
    compute_name: str = _empty
    description: str = ""
    units: str = _empty
    choices: set = _empty
    value: Any = _empty

    def update(self, other: dict) -> None:
        """Update a parameter's values from a dict."""
        for k, v in other.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                raise AttributeError(f"Unexpected parameter field '{k}'.")

    @classmethod
    def is_parameter_dict(cls, other: dict) -> bool:
        """Return whether other can update a parameter dictionary."""
        # Passing compute_name is forbidden.
        # name is valid, but is handled by the indicator
        return set(other.keys()).issubset(
            {"kind", "default", "description", "units", "choices", "value", "name"}
        )

    def __contains__(self, key) -> bool:
        """Imitate previous behaviour where "units" and "choices" were missing, instead of being "_empty"."""
        return getattr(self, key, _empty) is not _empty

    def asdict(self) -> dict:
        """Format indicators as a dictionary."""
        return {k: v for k, v in asdict(self).items() if v is not _empty}

    @property
    def injected(self) -> bool:
        """Indicate whether values are injected."""
        return self.value is not _empty


class IndicatorRegistrar:
    """Climate Indicator registering object."""

    def __new__(cls):
        """Add subclass to registry."""
        name = cls.__name__.upper()
        module = cls.__module__
        # If the module is not one of xclim's default, prepend the submodule name.
        if module.startswith("xclim.indicators"):
            submodule = module.split(".")[2]
            if submodule not in ["atmos", "generic", "land", "ocean", "seaIce"]:
                name = f"{submodule}.{name}"
        else:
            name = f"{module}.{name}"
        if name in registry:
            warnings.warn(
                f"Class {name} already exists and will be overwritten.", stacklevel=1
            )
        registry[name] = cls
        cls._registry_id = name
        return super().__new__(cls)

    def __init__(self):
        _indicators_registry[self.__class__].append(weakref.ref(self))

    @classmethod
    def get_instance(cls):
        """Return first found instance.

        Raises `ValueError` if no instance exists.
        """
        for inst_ref in _indicators_registry[cls]:
            inst = inst_ref()
            if inst is not None:
                return inst
        raise ValueError(
            f"There is no existing instance of {cls.__name__}. "
            "Either none were created or they were all garbage-collected."
        )


class Indicator(IndicatorRegistrar):
    r"""Climate indicator base class.

    Climate indicator object that, when called, computes an indicator and assigns its output a number of
    CF-compliant attributes. Some of these attributes can be *templated*, allowing metadata to reflect
    the value of call arguments.

    Instantiating a new indicator returns an instance but also creates and registers a custom subclass
    in :py:data:`xclim.core.indicator.registry`.

    Attributes in `Indicator.cf_attrs` will be formatted and added to the output variable(s).
    This attribute is a list of dictionaries. For convenience and retro-compatibility,
    standard CF attributes (names listed in :py:attr:`xclim.core.indicator.Indicator._cf_names`)
    can be passed as strings or list of strings directly to the indicator constructor.

    A lot of the Indicator's metadata is parsed from the underlying `compute` function's
    docstring and signature. Input variables and parameters are listed in
    :py:attr:`xclim.core.indicator.Indicator.parameters`, while parameters that will be
    injected in the compute function are in  :py:attr:`xclim.core.indicator.Indicator.injected_parameters`.
    Both are simply views of :py:attr:`xclim.core.indicator.Indicator._all_parameters`.

    Compared to their base `compute` function, indicators add the possibility of using dataset as input,
    with the added argument `ds` in the call signature. All arguments that were indicated
    by the compute function to be variables (DataArrays) through annotations will be promoted
    to also accept strings that correspond to variable names in the `ds` dataset.

    Parameters
    ----------
    identifier : str
        Unique ID for class registry, should be a valid slug.
    realm : {'atmos', 'seaIce', 'land', 'ocean'}
        General domain of validity of the indicator. Indicators created outside xclim.indicators must set this attribute.
    compute : func
        The function computing the indicators. It should return one or more DataArray.
    cf_attrs : list of dicts
        Attributes to be formatted and added to the computation's output.
        See :py:attr:`xclim.core.indicator.Indicator.cf_attrs`.
    title : str
        A succinct description of what is in the computed outputs. Parsed from `compute` docstring if None (first paragraph).
    abstract : str
        A long description of what is in the computed outputs. Parsed from `compute` docstring if None (second paragraph).
    keywords : str
        Comma separated list of keywords. Parsed from `compute` docstring if None (from a "Keywords" section).
    references : str
        Published or web-based references that describe the data or methods used to produce it. Parsed from
        `compute` docstring if None (from the "References" section).
    notes : str
        Notes regarding computing function, for example the mathematical formulation. Parsed from `compute`
        docstring if None (form the "Notes" section).
    src_freq : str, sequence of strings, optional
        The expected frequency of the input data. Can be a list for multiple frequencies, or None if irrelevant.
    context : str
        The `pint` unit context, for example use 'hydro' to allow conversion from kg m-2 s-1 to mm/day.

    Notes
    -----
    All subclasses created are available in the `registry` attribute and can be used to define custom subclasses
    or parse all available instances.
    """

    # Officially-supported metadata attributes on the output variables
    _cf_names = [
        "var_name",
        "standard_name",
        "long_name",
        "units",
        "units_metadata",
        "cell_methods",
        "description",
        "comment",
    ]

    # metadata fields that are formatted as free text (first letter capitalized)
    _text_fields = ["long_name", "description", "comment"]
    # Class attributes that are functions (so we know which to convert to static methods)
    _funcs = ["compute"]

    # Will become the class's name
    identifier = None

    context = "none"
    src_freq = None

    # Global metadata (must be strings, not attributed to the output)
    realm = None
    title = ""
    abstract = ""
    keywords = ""
    references = ""
    notes = ""
    _version_deprecated = ""

    # Note: typing and class types in this call signature will cause errors with sphinx-autodoc-typehints
    # See: https://github.com/tox-dev/sphinx-autodoc-typehints/issues/186#issuecomment-1450739378
    _all_parameters: dict = {}
    """A dictionary mapping metadata about the input parameters to the indicator.

    Keys are the arguments of the "compute" function. All parameters are listed, even
    those "injected", absent from the indicator's call signature. All are instances of
    :py:class:`xclim.core.indicator.Parameter`.
    """

    # Note: typing and class types in this call signature will cause errors with sphinx-autodoc-typehints
    # See: https://github.com/tox-dev/sphinx-autodoc-typehints/issues/186#issuecomment-1450739378
    cf_attrs: list[dict[str, str]] = None
    """A list of metadata information for each output of the indicator.

    It minimally contains a "var_name" entry, and may contain : "standard_name", "long_name",
    "units", "cell_methods", "description" and "comment" on official xclim indicators. Other
    fields could also be present if the indicator was created from outside xclim.

    var_name:
      Output variable(s) name(s). For derived single-output indicators, this field is not
      inherited from the parent indicator and defaults to the identifier.
    standard_name:
      Variable name, must be in the CF standard names table (this is not checked).
    long_name:
      Descriptive variable name. Parsed from `compute` docstring if not given.
      (first line after the output dtype, only works on single output function).
    units:
      Representative units of the physical quantity.
    cell_methods:
      List of blank-separated words of the form "name: method". Must respect the
      CF-conventions and vocabulary (not checked).
    description:
      Sentence(s) meant to clarify the qualifiers of the fundamental quantities, such as which
      surface a quantity is defined on or what the flux sign conventions are.
    comment:
      Miscellaneous information about the data or methods used to produce it.
    """

    def __new__(cls, **kwds):  # noqa: C901
        """Create subclass from arguments."""
        identifier = kwds.get("identifier", cls.identifier)
        if identifier is None:
            raise AttributeError("`identifier` has not been set.")

        if "compute" in kwds:
            # Parsed parameters and metadata override parent's params entirely.
            parameters, docmeta = cls._parse_indice(
                kwds["compute"], kwds.get("parameters", {})
            )
            for name, value in docmeta.items():
                # title, abstract, references, notes, long_name
                kwds.setdefault(name, value)

            # Added parameters
            # Subclasses can override or extend this through the classmethod _added_parameters
            for name, param in cls._added_parameters():
                if name in parameters:
                    raise ValueError(
                        f"Class {cls.__name__} can't wrap indices that have a `{name}`"
                        " argument as it conflicts with an argument it adds."
                    )
                parameters[name] = param
        else:  # inherit parameters from base class
            parameters = deepcopy(cls._all_parameters)

        # Update parameters with passed parameters, might change some parameters name (but not variables)
        cls._update_parameters(parameters, kwds.pop("parameters", {}))

        # Input variable mapping (to change variable names in signature and expected units/cf attrs).
        # new_units is a mapping from compute function name to units inferred from the var mapping
        new_units = cls._parse_var_mapping(kwds.pop("input", {}), parameters)

        # Raise on incorrect params, sort params, modify var defaults in-place if needed
        parameters = cls._ensure_correct_parameters(parameters)

        if "compute" in kwds and not hasattr(kwds["compute"], "in_units") and new_units:
            # If needed, wrap compute with declare units
            kwds["compute"] = declare_units(**new_units)(kwds["compute"])

        # All updates done.
        kwds["_all_parameters"] = parameters

        # Parse kwds to organize `cf_attrs`
        # And before converting callables to static methods
        kwds["cf_attrs"] = cls._parse_output_attrs(kwds, identifier)
        # Parse keywords
        if "keywords" in kwds:
            kwds["keywords"] = cls.keywords + " " + kwds.get("keywords")

        # Convert function objects to static methods.
        for key in cls._funcs:
            if key in kwds and callable(kwds[key]):
                kwds[key] = staticmethod(kwds[key])

        # Infer realm for built-in xclim instances
        if cls.__module__.startswith(__package__.split(".", maxsplit=1)[0]):
            xclim_realm = cls.__module__.split(".")[2]
        else:
            xclim_realm = None

        # Priority given to passed realm -> parent's realm -> location of the class declaration (official inds only)
        kwds.setdefault("realm", cls.realm or xclim_realm)
        if kwds["realm"] not in ["atmos", "seaIce", "land", "ocean", "generic"]:
            raise AttributeError(
                "Indicator's realm must be given as one of 'atmos', 'seaIce', 'land', 'ocean' or 'generic'"
            )

        # Create new class object
        new = type(identifier.upper(), (cls,), kwds)

        # Forcing the module is there so YAML-generated submodules are correctly seen by IndicatorRegistrar.
        if kwds.get("module") is not None:
            new.__module__ = f"xclim.indicators.{kwds['module']}"
        else:
            # If the module was not forced, set the module to the base class' module.
            # Otherwise, all indicators will have module `xclim.core.indicator`.
            new.__module__ = cls.__module__

        #  Add the created class to the registry
        # This will create an instance from the new class and call __init__.
        return super().__new__(new)

    @staticmethod
    def _parse_indice(compute, passed_parameters):  # noqa: F841
        """Parse the compute function.

        - Metadata is extracted from the docstring
        - Parameters are parsed from the docstring (description, choices), decorator (units), signature (kind, default)

        'passed_parameters' is only needed when compute is a generic function
        (not decorated by `declare_units`) and it takes a string parameter. In that case
        we need to check if that parameter has units (which have been passed explicitly).

        """
        docmeta = parse_doc(compute.__doc__)
        params_dict = docmeta.pop("parameters", {})  # override parent's parameters

        compute_sig = signature(compute)
        # Check that the `Parameters` section of the docstring does not include parameters
        # that are not in the `compute` function signature.
        if not set(params_dict.keys()).issubset(compute_sig.parameters.keys()):
            raise ValueError(
                f"Malformed docstring on {compute} : the parameters "
                f"{set(params_dict.keys()) - set(compute_sig.parameters.keys())} "
                "are absent from the signature."
            )
        for name, param in compute_sig.parameters.items():
            meta = params_dict.setdefault(name, {})
            meta["compute_name"] = name
            meta["default"] = param.default
            meta["kind"] = infer_kind_from_parameter(param)
            if hasattr(compute, "in_units") and name in compute.in_units:
                meta["units"] = compute.in_units[name]

        parameters = {name: Parameter(**param) for name, param in params_dict.items()}
        return parameters, docmeta

    @classmethod
    def _added_parameters(cls):
        """Create a list of tuples for arguments to add to the call signature (name, Parameter).

        These can't be in the compute function signature, the class is in charge of removing them
        from the params passed to the compute function, likely through an override of
        _preprocess_and_checks.
        """
        return [
            (
                "ds",
                Parameter(
                    kind=InputKind.DATASET,
                    default=None,
                    description="A dataset with the variables given by name.",
                ),
            )
        ]

    @classmethod
    def _update_parameters(cls, parameters, passed):
        """Update parameters with the ones passed."""
        try:
            for key, val in passed.items():
                if isinstance(val, dict) and Parameter.is_parameter_dict(val):
                    if "name" in val:
                        new_key = val.pop("name")
                        if new_key in parameters:
                            raise ValueError(
                                f"Cannot rename a parameter or variable with the same name as another parameter. '{new_key}' is already a parameter."
                            )
                        parameters[new_key] = parameters.pop(key)
                        key = new_key
                    parameters[key].update(val)
                elif key in parameters:
                    parameters[key].value = val
                else:
                    raise KeyError(key)
        except KeyError as err:
            raise ValueError(
                f"Parameter {err} was passed but it does not exist on the "
                f"compute function (not one of {parameters.keys()})"
            ) from err

    @classmethod
    def _parse_var_mapping(cls, variable_mapping, parameters):
        """Parse the variable mapping passed in `input` and update `parameters` in-place."""
        # Update parameters
        new_units = {}
        for old_name, new_name in variable_mapping.items():
            meta = parameters[new_name] = parameters.pop(old_name)
            try:
                varmeta = VARIABLES[new_name]
            except KeyError as err:
                raise ValueError(
                    f"Compute argument {old_name} was mapped to variable "
                    f"{new_name} which is not understood by xclim or CMIP6. Please"
                    " use names listed in `xclim.core.utils.VARIABLES`."
                ) from err
            if meta.units is not _empty:
                try:
                    check_units(varmeta["canonical_units"], meta.units)
                except ValidationError as err:
                    raise ValueError(
                        "When changing the name of a variable by passing `input`, "
                        "the units dimensionality must stay the same. Got: old = "
                        f"{meta.units}, new = {varmeta['canonical_units']}"
                    ) from err
            meta.units = varmeta.get("dimensions", varmeta["canonical_units"])
            new_units[meta.compute_name] = meta.units
            meta.description = varmeta["description"]
        return new_units

    @classmethod
    def _ensure_correct_parameters(cls, parameters):
        """Ensure the parameters are correctly set and ordered."""
        # Set default values, otherwise the signature binding chokes
        # on missing arguments when passing only `ds`.
        for name, meta in parameters.items():
            if not meta.injected:
                if meta.kind == InputKind.OPTIONAL_VARIABLE:
                    meta.default = None
                elif meta.kind in [InputKind.VARIABLE]:
                    meta.default = name

        # Sort parameters : Var, Opt Var, all params, ds, injected params.
        def sortkey(kv):
            if not kv[1].injected:
                if kv[1].kind in [
                    InputKind.VARIABLE,
                    InputKind.OPTIONAL_VARIABLE,
                    InputKind.KWARGS,
                ]:
                    return kv[1].kind
                return 2
            return 99

        return dict(sorted(parameters.items(), key=sortkey))

    @classmethod
    def _parse_output_attrs(  # noqa: C901
        cls, kwds: dict[str, Any], identifier: str
    ) -> list[dict[str, str | Callable]]:
        """CF-compliant metadata attributes for all output variables."""
        parent_cf_attrs = cls.cf_attrs
        cf_attrs = kwds.get("cf_attrs")
        if isinstance(cf_attrs, dict):
            # Single output indicator, but we store as a list anyway.
            cf_attrs = [cf_attrs]
        elif cf_attrs is None:
            # Attributes were passed the "old" way, with lists or strings directly (only _cf_names)
            # We need to get the number of outputs first, defaulting to the length of parent's cf_attrs or 1
            n_outs = len(parent_cf_attrs) if parent_cf_attrs is not None else 1
            for name in cls._cf_names:
                arg = kwds.get(name)
                if isinstance(arg, tuple | list):
                    n_outs = len(arg)

            # Populate new cf_attrs from parsing cf_names passed directly.
            cf_attrs = [{} for _ in range(n_outs)]
            for name in cls._cf_names:
                values = kwds.pop(name, None)
                if values is None:  # None passed, skip
                    continue
                if not isinstance(values, tuple | list):
                    # a single string or callable, same for all outputs
                    values = [values] * n_outs
                elif len(values) != n_outs:  # A sequence of the wrong length.
                    raise ValueError(
                        f"Attribute {name} has {len(values)} elements but xclim expected {n_outs}."
                    )
                for attrs, value in zip(cf_attrs, values, strict=False):
                    if value:  # Skip the empty ones (None or "")
                        attrs[name] = value
        # else we assume a list of dicts

        # For single output, var_name defaults to identifier.
        if len(cf_attrs) == 1 and "var_name" not in cf_attrs[0]:
            cf_attrs[0]["var_name"] = identifier

        # update from parent, if they have the same length.
        if parent_cf_attrs is not None and len(parent_cf_attrs) == len(cf_attrs):
            for old, new in zip(parent_cf_attrs, cf_attrs, strict=False):
                for attr, value in old.items():
                    new.setdefault(attr, value)

        # check if we have var_names for everybody
        for i, var in enumerate(cf_attrs, start=1):
            if "var_name" not in var:
                raise ValueError(f"Output #{i} is missing a var_name! Got: {var}.")

        return cf_attrs

    @classmethod
    def from_dict(
        cls,
        data: dict,
        identifier: str,
        module: str | None = None,
    ):
        """Create an indicator subclass and instance from a dictionary of parameters.

        Most parameters are passed directly as keyword arguments to the class constructor, except:

        - "base" : A subclass of Indicator or a name of one listed in
          :py:data:`xclim.core.indicator.registry` or
          :py:data:`xclim.core.indicator.base_registry`. When passed, it acts as if
          `from_dict` was called on that class instead.
        - "compute" : A string function name translates to a
          :py:mod:`xclim.indices.generic` or :py:mod:`xclim.indices` function.

        Parameters
        ----------
        data: dict
          The exact structure of this dictionary is detailed in the submodule documentation.
        identifier : str
          The name of the subclass and internal indicator name.
        module : str
          The module name of the indicator. This is meant to be used only if the indicator
          is part of a dynamically generated submodule, to override the module of the base class.
        """
        data = data.copy()
        if "base" in data:
            if isinstance(data["base"], str):
                parts = data["base"].split(".")
                registry_id = ".".join([*parts[:-1], parts[-1].upper()])
                cls = registry.get(registry_id, base_registry.get(data["base"]))
                if cls is None:
                    raise ValueError(
                        f"Requested base class {data['base']} is neither in the "
                        "indicators registry nor in base classes registry."
                    )
            else:
                cls = data["base"]

        compute = data.get("compute", None)
        # data.compute refers to a function in xclim.indices.generic or xclim.indices (in this order of priority).
        # It can also directly be a function (like if a module was passed to build_indicator_module_from_yaml)
        if isinstance(compute, str):
            compute_func = getattr(
                indices.generic, compute, getattr(indices, compute, None)
            )
            if compute_func is None:
                raise ImportError(
                    f"Indice function {compute} not found in xclim.indices or "
                    "xclim.indices.generic."
                )
            data["compute"] = compute_func

        return cls(identifier=identifier, module=module, **data)

    def __init__(self, **kwds):
        """Run checks and organizes the metadata."""
        # keywords of kwds that are class attributes have already been set in __new__
        self._check_identifier(self.identifier)

        # Validation is done : register the instance.
        super().__init__()

        self.__signature__ = self._gen_signature()

        # Generate docstring
        self.__doc__ = generate_indicator_docstring(self)

    def _gen_signature(self):
        """Generate the correct signature."""
        # Update call signature
        variables = []
        parameters = []
        compute_sig = signature(self.compute)
        for name, meta in self.parameters.items():
            if meta.kind in [
                InputKind.VARIABLE,
                InputKind.OPTIONAL_VARIABLE,
            ]:
                annot = DataArray | str
                if meta.kind == InputKind.OPTIONAL_VARIABLE:
                    annot = annot | None
                variables.append(
                    _Parameter(
                        name,
                        kind=_Parameter.POSITIONAL_OR_KEYWORD,
                        default=meta.default,
                        annotation=annot,
                    )
                )
            elif meta.kind == InputKind.KWARGS:
                parameters.append(_Parameter(name, kind=_Parameter.VAR_KEYWORD))
            elif meta.kind == InputKind.DATASET:
                parameters.append(
                    _Parameter(
                        name,
                        kind=_Parameter.KEYWORD_ONLY,
                        annotation=Dataset,
                        default=meta.default,
                    )
                )
            else:
                parameters.append(
                    _Parameter(
                        name,
                        kind=_Parameter.KEYWORD_ONLY,
                        default=meta.default,
                        annotation=compute_sig.parameters[meta.compute_name].annotation,
                    )
                )

        ret_ann = DataArray if self.n_outs == 1 else tuple[(DataArray,) * self.n_outs]
        return Signature(variables + parameters, return_annotation=ret_ann)

    def __call__(self, *args, **kwds):
        """Call function of Indicator class."""
        # Put the variables in `das`, parse them according to the following annotations:
        #     das : OrderedDict of variables (required + non-None optionals)
        #     params : OrderedDict of parameters (var_kwargs as a single argument, if any)

        if self._version_deprecated:
            self._show_deprecation_warning()  # noqa

        das, params, dsattrs = self._parse_variables_from_call(args, kwds)

        if OPTIONS[KEEP_ATTRS] is True or (
            OPTIONS[KEEP_ATTRS] == "xarray"
            and xarray.core.options._get_keep_attrs(False)
        ):
            out_attrs = xarray.core.merge.merge_attrs(
                [da.attrs for da in das.values()], "drop_conflicts"
            )
            out_attrs.pop("units", None)
        else:
            out_attrs = {}
        out_attrs = [out_attrs.copy() for i in range(self.n_outs)]

        das, params = self._preprocess_and_checks(das, params)

        # get mappings where keys are the actual compute function's argument names
        args = self._get_compute_args(das, params)
        with xarray.set_options(keep_attrs=False):
            outs = self.compute(**args)

        if isinstance(outs, DataArray):
            outs = [outs]

        if len(outs) != self.n_outs:
            raise ValueError(
                f"Indicator {self.identifier} was wrongly defined. Expected "
                f"{self.n_outs} outputs, got {len(outs)}."
            )

        # Metadata attributes from templates
        var_id = None
        for out, attrs, base_attrs in zip(outs, out_attrs, self.cf_attrs, strict=False):
            if self.n_outs > 1:
                var_id = base_attrs["var_name"]
            # Set default units
            attrs.update(units=out.units)
            if "units_metadata" in out.attrs:
                attrs["units_metadata"] = out.attrs["units_metadata"]
            attrs.update(
                self._update_attrs(
                    params.copy(),
                    das,
                    base_attrs,
                    names=self._cf_names,
                    var_id=var_id,
                )
            )

        # Convert to output units
        outs = [
            convert_units_to(out, attrs, self.context)
            for out, attrs in zip(outs, out_attrs, strict=False)
        ]

        outs = self._postprocess(outs, das, params)

        # Update variable attributes
        for out, attrs in zip(outs, out_attrs, strict=False):
            var_name = attrs.pop("var_name")
            out.attrs.update(attrs)
            out.name = var_name

        if OPTIONS[AS_DATASET]:
            out = Dataset({o.name: o for o in outs})
            if OPTIONS[KEEP_ATTRS] is True or (
                OPTIONS[KEEP_ATTRS] == "xarray"
                and xarray.core.options._get_keep_attrs(False)
            ):
                out.attrs.update(dsattrs)
            out.attrs["history"] = update_history(
                self._history_string(das, params),
                out,
                new_name=self.identifier,
            )
            return out

        # Return a single DataArray in case of single output, otherwise a tuple
        if self.n_outs == 1:
            return outs[0]
        return tuple(outs)

    def _parse_variables_from_call(self, args, kwds) -> tuple[OrderedDict, dict]:
        """Extract variable and optional variables from call arguments."""
        # Bind call arguments to `compute` arguments and set defaults.
        ba = self.__signature__.bind(*args, **kwds)
        ba.apply_defaults()

        # Assign inputs passed as strings from ds.
        self._assign_named_args(ba)

        # Extract variables + inject injected
        das = OrderedDict()
        params = ba.arguments.copy()
        for name, param in self._all_parameters.items():
            if not param.injected:
                # If a variable pop the arg
                if is_percentile_dataarray(params[name]):
                    # duplicate percentiles DA in both das and params
                    das[name] = params[name]
                elif param.kind in [InputKind.VARIABLE, InputKind.OPTIONAL_VARIABLE]:
                    data = params.pop(name)
                    # If a non-optional variable OR None, store the arg
                    if param.kind == InputKind.VARIABLE or data is not None:
                        das[name] = data
            else:
                params[name] = param.value

        ds = params.get("ds")
        dsattrs = ds.attrs if ds is not None else {}
        return das, params, dsattrs

    def _assign_named_args(self, ba):
        """Assign inputs passed as strings from ds."""
        ds = ba.arguments.get("ds")

        for name, val in ba.arguments.items():
            kind = self.parameters[name].kind

            if kind <= InputKind.OPTIONAL_VARIABLE:
                if isinstance(val, str) and ds is None:
                    raise ValueError(
                        "Passing variable names as string requires giving the `ds` "
                        f"dataset (got {name}='{val}')"
                    )
                if (isinstance(val, str) or val is None) and ds is not None:
                    # Set default name for DataArray
                    key = val or name

                    if key in ds:
                        ba.arguments[name] = ds[key]
                    elif kind == InputKind.VARIABLE:
                        raise MissingVariableError(
                            f"For input '{name}', variable '{key}' "
                            "was not found in the input dataset."
                        )

    def _preprocess_and_checks(self, das, params):
        """Actions to be done after parsing the arguments and before computing."""
        # Pre-computation validation checks on DataArray arguments
        self._bind_call(self.datacheck, **das)
        self._bind_call(self.cfcheck, **das)
        return das, params

    def _get_compute_args(self, das, params):
        """Rename variables and parameters to match the compute function's names and split VAR_KEYWORD arguments."""
        # Get correct variable names for the compute function.
        # Exclude param without a mapping inside the compute functions (those added by the indicator class)
        args = {}
        for key, p in self._all_parameters.items():
            if p.compute_name is not _empty:
                if key in das:
                    args[p.compute_name] = das[key]
                # elif because some args are in both (percentile DataArrays)
                elif key in params:
                    if p.kind == InputKind.KWARGS:
                        args.update(params[key])
                    else:
                        args[p.compute_name] = params[key]

        return args

    def _postprocess(self, outs, das, params):
        """Actions to done after computing."""
        return outs

    def _bind_call(self, func, **das):
        """Call function using `__call__` `DataArray` arguments.

        This will try to bind keyword arguments to `func` arguments. If this fails,
        `func` is called with positional arguments only.

        Notes
        -----
        This method is used to support two main use cases.

        In use case #1, we have two compute functions with arguments in a different order:
            `func1(tasmin, tasmax)` and `func2(tasmax, tasmin)`

        In use case #2, we have two compute functions with arguments that have different names:
            `generic_func(da)` and `custom_func(tas)`

        For each case, we want to define a single `cfcheck` and `datacheck` methods that
        will work with both compute functions.

        Passing a dictionary of arguments will solve #1, but not #2.
        """
        # First try to bind arguments to function.
        try:
            ba = signature(func).bind(**das)
        except TypeError:
            # If this fails, simply call the function using positional arguments
            return func(*das.values())
        # Call the func using bound arguments
        return func(*ba.args, **ba.kwargs)

    @classmethod
    def _get_translated_metadata(
        cls, locale, var_id=None, names=None, append_locale_name=True
    ):
        """Get raw translated metadata for the current indicator and a given locale.

        All available translated metadata from the current indicator and those it is
        based on are merged, with the highest priority set to the current one.
        """
        var_id = var_id or ""
        if var_id:
            var_id = "." + var_id

        family_tree = []
        cl = cls
        while hasattr(cl, "_registry_id"):
            family_tree.append(cl._registry_id + var_id)
            # The indicator mechanism always has single inheritance.
            cl = cl.__bases__[0]

        return get_local_attrs(
            family_tree,
            locale,
            names=names,
            append_locale_name=append_locale_name,
        )

    def _update_attrs(
        self,
        args: dict[str, Any],
        das: dict[str, DataArray],
        attrs: dict[str, str],
        var_id: str | None = None,
        names: Sequence[str] | None = None,
    ):
        """Format attributes with the run-time values of `compute` call parameters.

        Cell methods and history attributes are updated, adding to existing values.
        The language of the string is taken from the `OPTIONS` configuration dictionary.

        Parameters
        ----------
        args : dict[str, Any]
            Keyword arguments of the `compute` call.
        das : dict[str, DataArray]
            Input arrays.
        attrs : dict[str, str]
            The attributes to format and update.
        var_id : str
            The identifier to use when requesting the attributes translations.
            Defaults to the class name (for the translations) or the `identifier` field of
            the class (for the history attribute).
            If given, the identifier will be converted to uppercase to get the translation
            attributes. This is meant for multi-outputs indicators.
        names : sequence of str, optional
            List of attribute names for which to get a translation.

        Returns
        -------
        dict
            Attributes with {} expressions replaced by call argument values. With updated `cell_methods` and `history`.
            `cell_methods` is not added if `names` is given and those not contain `cell_methods`.
        """
        out = self._format(attrs, args)
        for locale in OPTIONS[METADATA_LOCALES]:
            out.update(
                self._format(
                    self._get_translated_metadata(
                        locale, var_id=var_id, names=names or list(attrs.keys())
                    ),
                    args=args,
                    formatter=get_local_formatter(locale),
                )
            )

        # Get history and cell method attributes from source data
        attrs = defaultdict(str)
        if names is None or "cell_methods" in names:
            attrs["cell_methods"] = merge_attributes(
                "cell_methods", new_line=" ", missing_str=None, **das
            )
            if "cell_methods" in out:
                attrs["cell_methods"] += " " + out.pop("cell_methods")

        attrs["history"] = update_history(
            self._history_string(das, args),
            new_name=out.get("var_name"),
            **das,
        )

        attrs.update(out)
        return attrs

    def _history_string(self, das, params):
        kwargs = {**das}
        for k, v in params.items():
            if self._all_parameters[k].injected:
                continue
            if self._all_parameters[k].kind == InputKind.KWARGS:
                kwargs.update(**v)
            elif self._all_parameters[k].kind != InputKind.DATASET:
                kwargs[k] = v
        return gen_call_string(self._registry_id, **kwargs)

    @staticmethod
    def _check_identifier(identifier: str) -> None:
        """Verify that the identifier is a proper slug."""
        if not re.match(r"^[-\w]+$", identifier):
            warnings.warn(
                "The identifier contains non-alphanumeric characters. "
                "It could make life difficult for downstream software reusing this class.",
                UserWarning,
            )

    @classmethod
    def translate_attrs(cls, locale: str | Sequence[str], fill_missing: bool = True):
        """Return a dictionary of unformatted translated translatable attributes.

        Translatable attributes are defined in :py:const:`xclim.core.locales.TRANSLATABLE_ATTRS`.

        Parameters
        ----------
        locale : str or sequence of str
            The POSIX name of the locale or a tuple of a locale name and a path to a json file defining translations.
            See `xclim.locale` for details.
        fill_missing : bool
            If True (default) fill the missing attributes by their english values.
        """

        def _translate(cf_attrs, names, var_id=None):
            attrs = cls._get_translated_metadata(
                locale,
                var_id=var_id,
                names=names,
                append_locale_name=False,
            )
            if fill_missing:
                for name in names:
                    if name not in attrs and cf_attrs.get(name):
                        attrs[name] = cf_attrs.get(name)
            return attrs

        # Translate global attrs
        attrs = _translate(
            cls.__dict__,
            # Translate only translatable attrs that are not variable attrs
            set(TRANSLATABLE_ATTRS).difference(set(cls._cf_names)),
        )
        # Translate variable attrs
        attrs["cf_attrs"] = []
        var_id = None
        for cf_attrs in cls.cf_attrs:  # Translate for each variable
            if len(cls.cf_attrs) > 1:
                var_id = cf_attrs["var_name"]
            attrs["cf_attrs"].append(
                _translate(
                    cf_attrs,
                    set(TRANSLATABLE_ATTRS).intersection(cls._cf_names),
                    var_id=var_id,
                )
            )
        return attrs

    @classmethod
    def json(cls, args=None):
        """Return a serializable dictionary representation of the class.

        Parameters
        ----------
        args : mapping, optional
            Arguments as passed to the call method of the indicator.
            If not given, the default arguments will be used when formatting the attributes.

        Notes
        -----
        This is meant to be used by a third-party library wanting to wrap this class into another interface.

        """
        names = ["identifier", "title", "abstract", "keywords"]
        out = {key: getattr(cls, key) for key in names}
        out = cls._format(out, args)

        # Format attributes
        out["outputs"] = [cls._format(attrs, args) for attrs in cls.cf_attrs]
        out["notes"] = cls.notes

        # We need to deepcopy, otherwise empty defaults get overwritten!
        # All those tweaks are to ensure proper serialization of the returned dictionary.
        out["parameters"] = {
            k: p.asdict() if not p.injected else deepcopy(p.value)
            for k, p in cls._all_parameters.items()
        }
        for name, param in list(out["parameters"].items()):
            if not cls._all_parameters[name].injected:
                param["kind"] = param["kind"].value  # Get the int.
                if "choices" in param:  # A set is stored, convert to list
                    param["choices"] = list(param["choices"])
                if param["default"] is _empty_default:
                    del param["default"]
            elif callable(param):  # Rare special case (doy_qmax and doy_qmin).
                out["parameters"][name] = f"{param.__module__}.{param.__name__}"

        return out

    @classmethod
    def _format(
        cls,
        attrs: dict,
        args: dict | None = None,
        formatter: AttrFormatter = default_formatter,
    ) -> dict:
        """Format attributes including {} tags with arguments.

        Parameters
        ----------
        attrs : dict
            Attributes containing tags to replace with arguments' values.
        args : dict, optional
            Function call arguments. If not given, the default arguments will be used when formatting the attributes.
        formatter : AttrFormatter
            Plaintext mappings for indicator attributes.

        Returns
        -------
        dict
        """
        # Use defaults
        if args is None:
            args = {
                k: p.default if not p.injected else p.value
                for k, p in cls._all_parameters.items()
            }

        # Prepare arguments
        mba = {}
        # Add formatting {} around values to be able to replace them with _attrs_mapping using format.
        for k, v in args.items():
            if isinstance(v, units.Quantity):
                mba[k] = f"{v:g~P}"
            elif isinstance(v, int | float):
                mba[k] = f"{v:g}"
            # TODO: What about InputKind.NUMBER_SEQUENCE
            elif k == "indexer":
                if v and v not in [_empty, _empty_default]:
                    dk, dv = v.copy().popitem()
                    if dk == "month":
                        dv = f"m{dv}"
                    elif dk in ("doy_bounds", "date_bounds"):
                        dv = f"{dv[0]} to {dv[1]}"
                    mba["indexer"] = dv
                else:
                    mba["indexer"] = args.get("freq") or "YS"
            elif is_percentile_dataarray(v):
                mba.update(get_percentile_metadata(v, k))
            elif (
                isinstance(v, DataArray)
                and cls._all_parameters[k].kind == InputKind.QUANTIFIED
            ):
                mba[k] = "<an array>"
            else:
                mba[k] = v
        out = {}
        for key, val in attrs.items():
            if callable(val):
                val = val(**mba)

            out[key] = formatter.format(val, **mba)

            if key in cls._text_fields:
                out[key] = out[key].strip().capitalize()

        return out

    # The following static methods are meant to be replaced to define custom indicators.
    @staticmethod
    def compute(*args, **kwds):
        """Compute the indicator.

        This would typically be a function from `xclim.indices`.
        """
        raise NotImplementedError

    def cfcheck(self, **das):
        """Compare metadata attributes to CF-Convention standards.

        Default cfchecks use the specifications in `xclim.core.utils.VARIABLES`,
        assuming the indicator's inputs are using the CMIP6/xclim variable names correctly.
        Variables absent from these default specs are silently ignored.

        When subclassing this method, use functions decorated using `xclim.core.options.cfcheck`.
        """
        for varname, vardata in das.items():
            try:
                cfcheck_from_name(varname, vardata)
            except KeyError:  # noqa: S110
                # Silently ignore unknown variables.
                pass

    def datacheck(self, **das):
        """Verify that input data is valid.

        When subclassing this method, use functions decorated using `xclim.core.options.datacheck`.

        For example, checks could include:

        * assert no precipitation is negative
        * assert no temperature has the same value 5 days in a row

        This base datacheck checks that the input data has a valid sampling frequency, as given in self.src_freq.
        If there are multiple inputs, it also checks if they all have the same frequency and the same anchor.
        """
        if self.src_freq is not None:
            for da in das.values():
                if "time" in da.coords and da.time.ndim == 1 and len(da.time) > 3:
                    datachecks.check_freq(da, self.src_freq, strict=True)

            datachecks.check_common_time(
                [
                    da
                    for da in das.values()
                    if "time" in da.coords and da.time.ndim == 1 and len(da.time) > 3
                ]
            )

    def __getattr__(self, attr):
        """Return the attribute."""
        if attr in self._cf_names:
            out = [meta.get(attr, "") for meta in self.cf_attrs]
            if len(out) == 1:
                return out[0]
            return out
        raise AttributeError(attr)

    @property
    def n_outs(self):
        """Return the length of all cf_attrs."""
        return len(self.cf_attrs)

    @property
    def parameters(self):
        """Create a dictionary of controllable parameters.

        Similar to :py:attr:`Indicator._all_parameters`, but doesn't include injected parameters.
        """
        return {
            name: param
            for name, param in self._all_parameters.items()
            if not param.injected
        }

    @property
    def injected_parameters(self):
        """Return a dictionary of all injected parameters.

        Opposite of :py:meth:`Indicator.parameters`.
        """
        return {
            name: param.value
            for name, param in self._all_parameters.items()
            if param.injected
        }

    @property
    def is_generic(self):
        """Return True if the indicator is "generic", meaning that it can accept variables with any units."""
        return not hasattr(self.compute, "in_units")

    def _show_deprecation_warning(self):
        warnings.warn(
            f"`{self.title}` is deprecated as of `xclim` v{self._version_deprecated} and will be removed "
            "in a future release. See the `xclim` release notes for more information: "
            f"https://xclim.readthedocs.io/en/stable/history.html",
            FutureWarning,
            stacklevel=3,
        )


class CheckMissingIndicator(Indicator):
    """Class adding missing value checks to indicators.

    This should not be used as-is, but subclassed by implementing the `_get_missing_freq` method.
    This method will be called in `_postprocess` using the compute parameters as only argument.
    It should return a freq string, the same as the output freq of the computed data.
    It can also be "None" to indicator the full time axis has been reduced, or "False" to skip the missing checks.

    Parameters
    ----------
    missing : {any, wmo, pct, at_least_n, skip, from_context}
        The name of the missing value method. See `xclim.core.missing.MissingBase` to create new custom methods.
        If None, this will be determined by the global configuration (see `xclim.set_options`).
        Defaults to "from_context".
    missing_options : dict, optional
        Arguments to pass to the `missing` function.
        If None, this will be determined by the global configuration.
    """

    missing = "from_context"
    missing_options: dict | None = None

    def __init__(self, **kwds):
        if self.missing == "from_context" and self.missing_options is not None:
            raise ValueError(
                "Cannot set `missing_options` with `missing` method being from context."
            )

        # Validate hard-coded missing options
        kls = MISSING_METHODS[self.missing]
        self._missing = kls.execute
        if self.missing_options:
            kls.validate(**self.missing_options)

        super().__init__(**kwds)

    def _history_string(self, das, params):
        if self.missing == "from_context":
            missing = OPTIONS[CHECK_MISSING]
        else:
            missing = self.missing
        opt_str = f" with options check_missing={missing}"

        if missing != "skip":
            mopts = self.missing_options or OPTIONS[MISSING_OPTIONS].get(missing)
            if mopts:
                opt_str += f", missing_options={mopts}"

        return super()._history_string(das, params) + opt_str

    def _get_missing_freq(self, params):
        """Return the resampling frequency to be used in the missing values check."""
        raise NotImplementedError("Don't use `CheckMissingIndicator` directly.")

    def _postprocess(self, outs, das, params):
        """Masking of missing values."""
        outs = super()._postprocess(outs, das, params)

        freq = self._get_missing_freq(params)
        if self.missing != "skip" or freq is False:
            # Mask results that do not meet criteria defined by the `missing` method.
            # This means all outputs must have the same dimensions as the broadcasted inputs (excluding time)
            options = self.missing_options or OPTIONS[MISSING_OPTIONS].get(
                self.missing, {}
            )

            # We flag periods according to the missing method. skip variables without a time coordinate.
            src_freq = self.src_freq if isinstance(self.src_freq, str) else None
            miss = (
                self._missing(da, freq, src_freq, options, params.get("indexer", {}))
                for da in das.values()
                if "time" in da.coords
            )
            # Reduce by or and broadcast to ensure the same length in time
            # When indexing is used and there are no valid points in the last period, mask will not include it
            mask = reduce(np.logical_or, miss)
            if isinstance(mask, DataArray):  # mask might be a bool in some cases
                if "time" in mask.dims and mask.time.size < outs[0].time.size:
                    mask = mask.reindex(time=outs[0].time, fill_value=True)
                # Remove any aux coord to avoid any unwanted dask computation in the alignment within "where"
                mask, _ = split_auxiliary_coordinates(mask)
            outs = [out.where(~mask) for out in outs]

        return outs


class ReducingIndicator(CheckMissingIndicator):
    """Indicator that performs a time-reducing computation.

    Compared to the base Indicator, this adds the handling of missing data.

    Parameters
    ----------
    missing : {any, wmo, pct, at_least_n, skip, from_context}
        The name of the missing value method. See `xclim.core.missing.MissingBase` to create new custom methods.
        If None, this will be determined by the global configuration (see `xclim.set_options`).
        Defaults to "from_context".
    missing_options : dict, optional
        Arguments to pass to the `missing` function.
        If None, this will be determined by the global configuration.
    """

    def _get_missing_freq(self, params):
        """Return None, to indicate that the full time axis is to be reduced."""
        return None


class ResamplingIndicator(CheckMissingIndicator):
    """Indicator that performs a resampling computation.

    Compared to the base Indicator, this adds the handling of missing data,
    and the check of allowed periods.

    Parameters
    ----------
    missing : {any, wmo, pct, at_least_n, skip, from_context}
        The name of the missing value method. See `xclim.core.missing.MissingBase` to create new custom methods.
        If None, this will be determined by the global configuration (see `xclim.set_options`).
        Defaults to "from_context".
    missing_options : dict, optional
        Arguments to pass to the `missing` function.
        If None, this will be determined by the global configuration.
    allowed_periods : Sequence[str], optional
        A list of allowed periods, i.e. base parts of the `freq` parameter.
        For example, indicators meant to be computed annually only will have `allowed_periods=["Y"]`.
        `None` means "any period" or that the indicator doesn't take a `freq` argument.
    """

    allowed_periods: list[str] | None = None

    @classmethod
    def _ensure_correct_parameters(cls, parameters):
        if "freq" not in parameters:
            raise ValueError(
                "ResamplingIndicator require a 'freq' argument, use the base Indicator"
                " class if your computation doesn't perform any resampling."
            )
        return super()._ensure_correct_parameters(parameters)

    def _get_missing_freq(self, params):
        return params["freq"]

    def _preprocess_and_checks(self, das, params):
        """Perform parent's checks and also check if freq is allowed."""
        das, params = super()._preprocess_and_checks(das, params)

        # Check if the period is allowed:
        if self.allowed_periods is not None:
            if parse_offset(params["freq"])[1] not in self.allowed_periods:
                raise ValueError(
                    f"Resampling frequency {params['freq']} is not allowed for indicator "
                    f"{self.identifier} (needs something equivalent to one "
                    f"of {self.allowed_periods})."
                )

        return das, params


class IndexingIndicator(Indicator):
    """Indicator that also adds the "indexer" kwargs to subset the inputs before computation."""

    @classmethod
    def _added_parameters(cls):
        """Create a list of tuples for arguments to add (name, Parameter)."""
        return super()._added_parameters() + [
            (
                "indexer",
                Parameter(
                    kind=InputKind.KWARGS,
                    description=(
                        "Indexing parameters to compute the indicator on a temporal "
                        "subset of the data. It accepts the same arguments as "
                        ":py:func:`xclim.indices.generic.select_time`."
                    ),
                ),
            )
        ]

    def _preprocess_and_checks(self, das: dict[str, DataArray], params: dict[str, Any]):
        """Perform parent's checks and also check if freq is allowed."""
        das, params = super()._preprocess_and_checks(das, params)

        indxr = params.get("indexer")
        if indxr:
            for k, da in filter(lambda kda: "time" in kda[1].coords, das.items()):
                das.update({k: select_time(da, **indxr)})
        return das, params


class ResamplingIndicatorWithIndexing(ResamplingIndicator, IndexingIndicator):
    """Resampling indicator that also adds "indexer" kwargs to subset the inputs before computation."""


class Daily(ResamplingIndicator):
    """Class for daily inputs and resampling computes."""

    src_freq = "D"


class Hourly(ResamplingIndicator):
    """Class for hourly inputs and resampling computes."""

    src_freq = "h"


base_registry["Indicator"] = Indicator
base_registry["ReducingIndicator"] = ReducingIndicator
base_registry["IndexingIndicator"] = IndexingIndicator
base_registry["ResamplingIndicator"] = ResamplingIndicator
base_registry["ResamplingIndicatorWithIndexing"] = ResamplingIndicatorWithIndexing
base_registry["Hourly"] = Hourly
base_registry["Daily"] = Daily


def add_iter_indicators(module):
    """Create an iterable of loaded indicators."""
    if not hasattr(module, "iter_indicators"):

        def iter_indicators():
            for ind_name, ind in module.__dict__.items():
                if isinstance(ind, Indicator):
                    yield ind_name, ind

        iter_indicators.__doc__ = f"Iterate over the (name, indicator) pairs in the {module.__name__} indicator module."

        module.__dict__["iter_indicators"] = iter_indicators


def build_indicator_module(
    name: str,
    objs: dict[str, Indicator],
    doc: str | None = None,
    reload: bool = False,
) -> ModuleType:
    """Create or update a module from imported objects.

    The module is inserted as a submodule of :py:mod:`xclim.indicators`.

    Parameters
    ----------
    name : str
        New module name.
        If it already exists, the module is extended with the passed objects, overwriting those with same names.
    objs : dict[str, Indicator]
        Mapping of the indicators to put in the new module.
        Keyed by the name they will take in that module.
    doc : str
        Docstring of the new module. Defaults to a simple header.
        Invalid if the module already exists.
    reload : bool
        If reload is True and the module already exists, it is first removed before being rebuilt.
        If False (default), indicators are added or updated, but not removed.

    Returns
    -------
    ModuleType
        A indicator module built from a mapping of Indicators.
    """
    from xclim import indicators  # pylint: disable=import-outside-toplevel

    out: ModuleType
    if hasattr(indicators, name):
        if doc is not None:
            warnings.warn(
                "Passed docstring ignored when extending existing module.", stacklevel=1
            )
        out = getattr(indicators, name)
        if reload:
            for n, ind in list(out.iter_indicators()):
                if n not in objs:
                    # Remove the indicator from the registries and the module
                    del registry[ind._registry_id]  # noqa
                    del _indicators_registry[ind.__class__]
                    del out.__dict__[n]
    else:
        doc = doc or f"{name.capitalize()} indicators\n" + "=" * (len(name) + 11)
        try:
            out = ModuleType(name, doc)
        except TypeError as err:
            raise TypeError(f"Module '{name}' is not properly formatted") from err
        indicators.__dict__[name] = out

    out.__dict__.update(objs)
    add_iter_indicators(out)
    return out


def build_indicator_module_from_yaml(  # noqa: C901
    filename: PathLike,
    name: str | None = None,
    indices: dict[str, Callable] | ModuleType | PathLike | None = None,
    translations: dict[str, dict | PathLike] | None = None,
    mode: str = "raise",
    encoding: str = "UTF8",
    reload: bool = False,
    validate: bool | PathLike = True,
) -> ModuleType:
    """Build or extend an indicator module from a YAML file.

    The module is inserted as a submodule of :py:mod:`xclim.indicators`.
    When given only a base filename (no 'yml' extension), this tries to find custom indices in a module
    of the same name (*.py) and translations in json files (*.<lang>.json), see Notes.

    Parameters
    ----------
    filename : PathLike
        Path to a YAML file or to the stem of all module files. See Notes for behaviour when passing a basename only.
    name : str, optional
        The name of the new or existing module, defaults to the basename of the file.
        (e.g: `atmos.yml` -> `atmos`)
    indices : Mapping of callables or module or path, optional
        A mapping or module of indice functions or a python file declaring such a file.
        When creating the indicator, the name in the `index_function` field is first sought
        here, then the indicator class will search in xclim.indices.generic and finally in xclim.indices.
    translations : Mapping of dicts or path, optional
        Translated metadata for the new indicators. Keys of the mapping must be 2-char language tags.
        Values can be translations dictionaries as defined in :ref:`internationalization:Internationalization`.
        They can also be a path to a json file defining the translations.
    mode : {'raise', 'warn', 'ignore'}
        How to deal with broken indice definitions.
    encoding : str
        The encoding used to open the `.yaml` and `.json` files.
        It defaults to UTF-8, overriding python's mechanism which is machine dependent.
    reload : bool
        If reload is True and the module already exists, it is first removed before being rebuilt.
        If False (default), indicators are added or updated, but not removed.
    validate : bool or path
        If True (default), the yaml module is validated against xclim's schema.
        Can also be the path to a yml schema against which to validate.
        Or False, in which case validation is simply skipped.

    Returns
    -------
    ModuleType
        A submodule of `pym:mod:`xclim.indicators`.

    Notes
    -----
    When the given `filename` has no suffix (usually '.yaml' or '.yml'), the function will try to load
    custom indice definitions from a file with the same name but with a `.py` extension. Similarly,
    it will try to load translations in `*.<lang>.json` files, where `<lang>` is the IETF language tag.

    For example. a set of custom indicators could be fully described by the following files:

        - `example.yml` : defining the indicator's metadata.
        - `example.py` : defining a few indice functions.
        - `example.fr.json` : French translations
        - `example.tlh.json` : Klingon translations.

    See Also
    --------
    xclim.core.indicator

    build_module
    """
    filepath = Path(filename)

    if not filepath.suffix:
        # A stem was passed, try to load files
        ymlpath = filepath.with_suffix(".yml")
    else:
        ymlpath = filepath

    # Read YAML file
    with ymlpath.open(encoding=encoding) as f:
        yml = safe_load(f)

    if validate is not False:
        # Read schema
        if validate is not True:
            schema = yamale.make_schema(validate)
        else:
            schema = yamale.make_schema(
                Path(__file__).parent.parent / "data" / "schema.yml"
            )

        # Validate - a YamaleError will be raised if the module does not comply with the schema.
        yamale.validate(
            schema, yamale.make_data(content=ymlpath.read_text(encoding=encoding))
        )

    # Load values from top-level in yml.
    # Priority of arguments differ.
    module_name = name or yml.get("module", filepath.stem)
    default_base = registry.get(
        yml.get("base"), base_registry.get(yml.get("base"), Daily)
    )
    doc = yml.get("doc")

    if (
        not filepath.suffix
        and indices is None
        and (indfile := filepath.with_suffix(".py")).is_file()
    ):
        # No suffix means we try to automatically detect the python file
        indices = indfile

    if isinstance(indices, str | Path):
        indices = load_module(indices, name=module_name)

    _translations: dict[str, dict] = {}
    if not filepath.suffix and translations is None:
        # No suffix mean we try to automatically detect the json files.
        for locfile in filepath.parent.glob(f"{filepath.stem}.*.json"):
            locale = locfile.suffixes[0][1:]
            _translations[locale] = read_locale_file(
                locfile, module=module_name, encoding=encoding
            )
    elif translations is not None:
        # A mapping was passed, we read paths is any.
        _translations = {
            lng: (
                read_locale_file(trans, module=module_name, encoding=encoding)
                if isinstance(trans, str | Path)
                else trans
            )
            for lng, trans in translations.items()
        }

    # Module-wide default values for some attributes
    defkwargs = {
        # Only used in case the indicator definition does not give them.
        "realm": yml.get("realm", "atmos"),
        # Merged with a space
        "keywords": yml.get("keywords"),
        # Merged with a new line
        "references": yml.get("references"),
    }

    def _merge_attrs(dbase, dextra, attr, sep):
        """Merge or replace attribute in dbase from dextra."""
        a = dbase.get(attr)
        b = dextra.get(attr)
        # If both are not None, join.
        if a and b:
            dbase[attr] = sep.join([a, b])
        # Otherwise, if a is None, but not b, replace.
        elif b:
            dbase[attr] = b

    # Parse the variables:
    for varname, vardata in yml.get("variables", {}).items():
        if varname in VARIABLES and VARIABLES[varname] != vardata:
            warnings.warn(
                f"Variable {varname} from module {module_name} "
                "will overwrite the one already defined in `xclim.core.utils.VARIABLES`"
            )
        VARIABLES[varname] = vardata.copy()

    # Parse the indicators:
    mapping = {}
    for identifier, data in yml["indicators"].items():
        try:
            # Get base class if it was relative to this module
            if "base" in data:
                if data["base"].startswith("."):
                    # A point means the base has been declared above.
                    data["base"] = registry[module_name + data["base"].upper()]
            else:
                # If no base is specified, pass the default one.
                data["base"] = default_base

            # Get the compute function if it is from the passed mapping
            if indices is not None and "compute" in data:
                indice_name = data["compute"]
                indice_func = getattr(indices, indice_name, None)
                if indice_func is None and hasattr(indices, "__getitem__"):
                    try:
                        indice_func = indices[indice_name]
                    except KeyError:  # noqa: S110
                        # The indice is not found in the mapping.
                        pass

                if indice_func is not None:
                    data["compute"] = indice_func

            _merge_attrs(data, defkwargs, "references", "\n")
            _merge_attrs(data, defkwargs, "keywords", " ")
            if data.get("realm") is None and defkwargs.get("realm") is not None:
                data["realm"] = defkwargs["realm"]

            mapping[identifier] = Indicator.from_dict(
                data, identifier=identifier, module=module_name
            )

        except Exception as err:  # pylint: disable=broad-except
            raise_warn_or_log(
                err, mode, msg=f"Constructing {identifier} failed with {err!r}"
            )

    # Construct module
    mod = build_indicator_module(module_name, objs=mapping, doc=doc, reload=reload)

    # If there are translations, load them
    if _translations:
        for locale, loc_dict in _translations.items():
            load_locale(loc_dict, locale)

    return mod



"""
Ensembles Creation and Statistics
=================================
"""

from __future__ import annotations

from collections.abc import Sequence
from glob import glob
from pathlib import Path
from typing import Any, Literal

import numpy as np
import xarray as xr

from xclim.core.calendar import common_calendar, get_calendar
from xclim.core.formatting import update_history
from xclim.core.utils import calc_perc

# The alpha and beta parameters for the quantile function
_quantile_params = {
    "interpolated_inverted_cdf": (0, 1),
    "hazen": (0.5, 0.5),
    "weibull": (0, 0),
    "linear": (1, 1),
    "median_unbiased": (1 / 3, 1 / 3),
    "normal_unbiased": (3 / 8, 3 / 8),
}


def create_ensemble(
    datasets: Any,
    multifile: bool = False,
    resample_freq: str | None = None,
    calendar: str | None = None,
    realizations: Sequence[Any] | None = None,
    cal_kwargs: dict | None = None,
    **xr_kwargs,
) -> xr.Dataset:
    r"""Create an xarray dataset of an ensemble of climate simulation from a list of netcdf files.

    Input data is concatenated along a newly created data dimension ('realization'). Returns an xarray dataset object
    containing input data from the list of netcdf files concatenated along a new dimension (name:'realization').
    In the case where input files have unequal time dimensions, the output ensemble Dataset is created for maximum
    time-step interval of all input files.  Before concatenation, datasets not covering the entire time span have
    their data padded with NaN values. Dataset and variable attributes of the first dataset are copied to the
    resulting dataset.

    Parameters
    ----------
    datasets : list or dict or string
        List of netcdf file paths or xarray Dataset/DataArray objects . If `multifile` is True, ncfiles should be a
        list of lists where each sublist contains input .nc files of an xarray multifile Dataset.
        If DataArray objects are passed, they should have a name in order to be transformed into Datasets.
        A dictionary can be passed instead of a list, in which case the keys are used as coordinates along the new
        `realization` axis.
        If a string is passed, it is assumed to be a glob pattern for finding datasets.
    multifile : bool
        If True, climate simulations are treated as xarray multifile Datasets before concatenation.
        Only applicable when "datasets" is sequence of list of file paths. Default: False.
    resample_freq : Optional[str]
        If the members of the ensemble have the same frequency but not the same offset, they cannot be properly aligned.
        If resample_freq is set, the time coordinate of each member will be modified to fit this frequency.
    calendar : str, optional
        The calendar of the time coordinate of the ensemble.
        By default, the smallest common calendar is chosen. For example, a mixed input of "noleap" and "360_day" will default to "noleap".
        'default' is the standard calendar using np.datetime64 objects (xarray's "standard" with `use_cftime=False`).
    realizations : sequence, optional
        The coordinate values for the new `realization` axis.
        If None (default), the new axis has a simple integer coordinate.
        This argument shouldn't be used if `datasets` is a glob pattern as the dataset order is random.
    cal_kwargs : dict, optional
        Additional arguments to pass to py:func:`xclim.core.calendar.convert_calendar`.
        For conversions involving '360_day', the align_on='date' option is used by default.
    \*\*xr_kwargs
        Any keyword arguments to be given to `xr.open_dataset` when opening the files
        (or to `xr.open_mfdataset` if `multifile` is True)

    Returns
    -------
    xr.Dataset
        Dataset containing concatenated data from all input files.

    Notes
    -----
    Input netcdf files require equal spatial dimension size (e.g. lon, lat dimensions).
    If input data contains multiple cftime calendar types they must be at monthly or coarser frequency.

    Examples
    --------
    .. code-block:: python

        from pathlib import Path
        from xclim.ensembles import create_ensemble

        ens = create_ensemble(temperature_datasets)

        # Using multifile datasets, through glob patterns.
        # Simulation 1 is a list of .nc files (e.g. separated by time):
        datasets = list(Path("/dir").glob("*.nc"))

        # Simulation 2 is also a list of .nc files:
        datasets.extend(Path("/dir2").glob("*.nc"))
        ens = create_ensemble(datasets, multifile=True)
    """
    if isinstance(datasets, dict):
        if realizations is None:
            realizations, datasets = zip(*datasets.items(), strict=False)
        else:
            datasets = datasets.values()
    elif isinstance(datasets, str) and realizations is not None:
        raise ValueError(
            "Passing `realizations` is not supported when `datasets` "
            "is a glob pattern, as the final order is random."
        )

    ds = _ens_align_datasets(
        datasets,
        multifile,
        resample_freq,
        calendar=calendar,
        cal_kwargs=cal_kwargs or {},
        **xr_kwargs,
    )

    if realizations is None:
        realizations = np.arange(len(ds))

    dim = xr.IndexVariable("realization", list(realizations), attrs={"axis": "E"})

    ens = xr.concat(ds, dim)
    for var_name, var in ds[0].variables.items():
        ens[var_name].attrs.update(**var.attrs)
    ens.attrs.update(**ds[0].attrs)

    return ens


def ensemble_mean_std_max_min(
    ens: xr.Dataset, min_members: int | None = 1, weights: xr.DataArray | None = None
) -> xr.Dataset:
    """Calculate ensemble statistics between a results from an ensemble of climate simulations.

    Returns an xarray Dataset containing ensemble mean, standard-deviation, minimum and maximum for input climate
    simulations.

    Parameters
    ----------
    ens : xr.Dataset
        Ensemble dataset (see xclim.ensembles.create_ensemble).
    min_members : int, optional
        The minimum number of valid ensemble members for a statistic to be valid.
        Passing None is equivalent to setting min_members to the size of the realization dimension.
        The default (1) essentially skips this check.
    weights : xr.DataArray, optional
        Weights to apply along the 'realization' dimension. This array cannot contain missing values.

    Returns
    -------
    xr.Dataset
        Dataset with data variables of ensemble statistics.

    Examples
    --------
    .. code-block:: python

        from xclim.ensembles import create_ensemble, ensemble_mean_std_max_min

        # Create the ensemble dataset:
        ens = create_ensemble(temperature_datasets)

        # Calculate ensemble statistics:
        ens_mean_std = ensemble_mean_std_max_min(ens)
    """
    if min_members is None:
        min_members = ens.realization.size
    ds_out = xr.Dataset(attrs=ens.attrs)
    for v in ens.data_vars:
        if weights is None:
            ds_out[f"{v}_mean"] = ens[v].mean(dim="realization")
            ds_out[f"{v}_stdev"] = ens[v].std(dim="realization")
        else:
            with xr.set_options(keep_attrs=True):
                ds_out[f"{v}_mean"] = ens[v].weighted(weights).mean(dim="realization")
                ds_out[f"{v}_stdev"] = ens[v].weighted(weights).std(dim="realization")
        ds_out[f"{v}_max"] = ens[v].max(dim="realization")
        ds_out[f"{v}_min"] = ens[v].min(dim="realization")

        if min_members != 1:
            enough = ens[v].notnull().sum("realization") >= min_members

        # Re-add attributes
        for stat in ["mean", "stdev", "max", "min"]:
            vv = f"{v}_{stat}"
            if min_members != 1:
                ds_out[vv] = ds_out[vv].where(enough)
            ds_out[vv].attrs = ens[v].attrs
            if "description" in ds_out[vv].attrs.keys():
                vv.split()
                ds_out[vv].attrs["description"] = (
                    ds_out[vv].attrs["description"]
                    + " : "
                    + vv.split("_")[-1]
                    + " of ensemble"
                )

    ds_out.attrs["history"] = update_history(
        f"Computation of statistics on {ens.realization.size} ensemble members.", ds_out
    )
    return ds_out


def ensemble_percentiles(
    ens: xr.Dataset | xr.DataArray,
    values: Sequence[int] | None = None,
    keep_chunk_size: bool | None = None,
    min_members: int | None = 1,
    weights: xr.DataArray | None = None,
    split: bool = True,
    method: Literal[
        "linear",
        "interpolated_inverted_cdf",
        "hazen",
        "weibull",
        "median_unbiased",
        "normal_unbiased",
    ] = "linear",
) -> xr.DataArray | xr.Dataset:
    """Calculate ensemble statistics between a results from an ensemble of climate simulations.

    Returns a Dataset containing ensemble percentiles for input climate simulations.

    Parameters
    ----------
    ens : xr.Dataset or xr.DataArray
        Ensemble Dataset or DataArray (see xclim.ensembles.create_ensemble).
    values : Sequence[int], optional
        Percentile values to calculate. Default: (10, 50, 90).
    keep_chunk_size : bool, optional
        For ensembles using dask arrays, all chunks along the 'realization' axis are merged.
        If True, the dataset is rechunked along the dimension with the largest chunks,
        so that the chunks keep the same size (approximately).
        If False, no shrinking is performed, resulting in much larger chunks.
        If not defined, the function decides which is best.
    min_members : int, optional
        The minimum number of valid ensemble members for a statistic to be valid.
        Passing None is equivalent to setting min_members to the size of the realization dimension.
        The default (1) essentially skips this check.
    weights : xr.DataArray, optional
        Weights to apply along the 'realization' dimension. This array cannot contain missing values.
        When given, the function uses xarray's quantile method which is slower than xclim's NaN-optimized algorithm,
        and does not support `method` values other than `linear`.
    split : bool
        Whether to split each percentile into a new variable
        or concatenate the output along a new "percentiles" dimension.
    method : {"linear", "interpolated_inverted_cdf", "hazen", "weibull", "median_unbiased", "normal_unbiased"}
        Method to use for estimating the percentile, see the `numpy.percentile` documentation for more information.

    Returns
    -------
    xr.Dataset or xr.DataArray
        If split is True, same type as ens; dataset otherwise,
        containing data variable(s) of requested ensemble statistics

    Examples
    --------
    .. code-block:: python

        from xclim.ensembles import create_ensemble, ensemble_percentiles

        # Create ensemble dataset:
        ens = create_ensemble(temperature_datasets)

        # Calculate default ensemble percentiles:
        ens_percs = ensemble_percentiles(ens)

        # Calculate non-default percentiles (25th and 75th)
        ens_percs = ensemble_percentiles(ens, values=(25, 50, 75))

        # If the original array has many small chunks, it might be more efficient to do:
        ens_percs = ensemble_percentiles(ens, keep_chunk_size=False)
    """
    if values is None:
        values = [10, 50, 90]
    if min_members is None:
        min_members = ens.realization.size

    if isinstance(ens, xr.Dataset):
        out = xr.merge(
            [
                ensemble_percentiles(
                    da,
                    values,
                    keep_chunk_size=keep_chunk_size,
                    split=split,
                    min_members=min_members,
                    weights=weights,
                    method=method,
                )
                for da in ens.data_vars.values()
                if "realization" in da.dims
            ]
        )
        out.attrs.update(ens.attrs)
        out.attrs["history"] = update_history(
            f"Computation of the percentiles on {ens.realization.size} ensemble members.",
            ens,
        )

        return out

    # Percentile calculation forbids any chunks along realization
    if ens.chunks and len(ens.chunks[ens.get_axis_num("realization")]) > 1:
        if keep_chunk_size is None:
            # Enable smart rechunking is chunksize exceed 2E8 elements after merging along realization
            keep_chunk_size = (
                np.prod(ens.isel(realization=0).data.chunksize) * ens.realization.size
                > 2e8
            )
        if keep_chunk_size:
            # Smart rechunk on dimension where chunks are the largest
            chk_dim, chks = max(
                enumerate(ens.chunks),
                key=lambda kv: (
                    0 if kv[0] == ens.get_axis_num("realization") else max(kv[1])
                ),
            )
            ens = ens.chunk(
                {"realization": -1, ens.dims[chk_dim]: len(chks) * ens.realization.size}
            )
        else:
            ens = ens.chunk({"realization": -1})

    if weights is None:
        alpha, beta = _quantile_params[method]

        out = xr.apply_ufunc(
            calc_perc,
            ens,
            input_core_dims=[["realization"]],
            output_core_dims=[["percentiles"]],
            keep_attrs=True,
            kwargs={"percentiles": values, "alpha": alpha, "beta": beta},
            dask="parallelized",
            output_dtypes=[ens.dtype],
            dask_gufunc_kwargs={"output_sizes": {"percentiles": len(values)}},
        )
    else:
        if method != "linear":
            raise ValueError(
                "Only the 'linear' method is supported when using weights."
            )

        with xr.set_options(keep_attrs=True):
            # xclim's calc_perc does not support weighted arrays, so xarray's native function is used instead.
            qt = np.array(values) / 100  # xarray requires values between 0 and 1
            out = (
                ens.weighted(weights)
                .quantile(qt, dim="realization", keep_attrs=True)
                .rename({"quantile": "percentiles"})
            )

    if min_members != 1:
        out = out.where(ens.notnull().sum("realization") >= min_members)
    out = out.assign_coords(
        percentiles=xr.DataArray(list(values), dims=("percentiles",))
    )

    if split:
        out = out.to_dataset(dim="percentiles")
        for p, perc in out.data_vars.items():
            perc.attrs.update(ens.attrs)
            perc.attrs["description"] = (
                perc.attrs.get("description", "") + f" {p}th percentile of ensemble."
            )
            out[p] = perc
            out = out.rename(name_dict={p: f"{ens.name}_p{int(p):02d}"})

    out.attrs["history"] = update_history(
        f"Computation of the percentiles on {ens.realization.size} ensemble members using method {method}.",
        ens,
    )

    return out


def _ens_align_datasets(
    datasets: list[xr.Dataset | Path | str | list[Path | str]] | str,
    multifile: bool = False,
    resample_freq: str | None = None,
    calendar: str = "default",
    cal_kwargs: dict | None = None,
    **xr_kwargs,
) -> list[xr.Dataset]:
    r"""Create a list of aligned xarray Datasets for ensemble Dataset creation.

    Parameters
    ----------
    datasets : list[xr.Dataset | xr.DataArray | Path | str | list[Path | str]] or str
        List of netcdf file paths or xarray Dataset/DataArray objects . If `multifile` is True, 'datasets' should be a
        list of lists where each sublist contains input NetCDF files of a xarray multi-file Dataset.
        DataArrays should have a name, so they can be converted to datasets.
        If a string, it is assumed to be a glob pattern for finding datasets.
    multifile : bool
        If True climate simulations are treated as xarray multi-file datasets before concatenation.
        Only applicable when 'datasets' is a sequence of file paths.
    resample_freq : str, optional
        If the members of the ensemble have the same frequency but not the same offset, they cannot be properly aligned.
        If resample_freq is set, the time coordinate of each member will be modified to fit this frequency.
    calendar : str
        The calendar of the time coordinate of the ensemble. For conversions involving '360_day',
        the align_on='date' option is used.
        See :py:func:`xclim.core.calendar.convert_calendar`.
        'default' is the standard calendar using np.datetime64 objects.
    \*\*xr_kwargs
        Any keyword arguments to be given to xarray when opening the files.

    Returns
    -------
    list[xr.Dataset]
    """
    xr_kwargs.setdefault("chunks", "auto")
    xr_kwargs.setdefault("decode_times", False)

    if isinstance(datasets, str):
        datasets = glob(datasets)

    ds_all: list[xr.Dataset] = []
    calendars = []
    for i, n in enumerate(datasets):
        ds: xr.Dataset
        if multifile:
            ds = xr.open_mfdataset(n, combine="by_coords", **xr_kwargs)
        else:
            if isinstance(n, xr.Dataset):
                ds = n
            elif isinstance(n, xr.DataArray):
                ds = n.to_dataset()
            else:
                ds = xr.open_dataset(n, **xr_kwargs)

        if "time" in ds.coords:
            time = xr.decode_cf(ds).time

            if resample_freq is not None:
                # Cast to bool to avoid bug in flox/numpy_groupies (xarray-contrib/flox#137)
                counts = time.astype(bool).resample(time=resample_freq).count()
                if any(counts > 1):
                    raise ValueError(
                        f"Alignment of dataset #{i:02d} failed: "
                        f"Time axis cannot be resampled to freq {resample_freq}."
                    )
                time = counts.time

            ds["time"] = time
            calendars.append(get_calendar(time))

        ds_all.append(ds)

    if not calendars:
        # no time
        return ds_all

    if calendar is None:
        calendar = common_calendar(calendars, join="outer")
    cal_kwargs.setdefault("align_on", "date")
    return [ds.convert_calendar(calendar, **cal_kwargs) for ds in ds_all]



from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr

import xclim.indices as xci
from xclim import atmos, core, set_options
from xclim.core.calendar import build_climatology_bounds, percentile_doy
from xclim.core.units import convert_units_to

K2C = 273.15


class TestRainOnFrozenGround:
    @pytest.mark.parametrize("chunks", [{"time": 366}, None])
    def test_3d_data_with_nans(self, open_dataset, chunks):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")

        pr = ds.pr.copy()
        pr.values[1, 10] = np.nan

        if chunks:
            ds = ds.chunk(chunks)

        out = atmos.rain_on_frozen_ground_days(pr, ds.tas, freq="YS")
        np.testing.assert_array_equal(out.sel(location="Montréal"), [np.nan, 4, 5, 3])


class TestRainSeason:
    # @pytest.mark.parametrize("chunks", [{"time": 366}, None])
    def test_3d_data_with_nans(self, open_dataset):
        ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")

        pr = ds.pr.isel(location=0).copy()
        pr[{"time": [0, 10, 100]}] = np.nan
        out = {}
        out["start"], out["end"], out["length"] = atmos.rain_season(
            pr,
            freq="YS-JAN",
            window_dry_end=5,
            date_min_start="01-01",
            date_min_end="01-01",
        )
        out_arr = np.array([out[var].values for var in ["start", "end", "length"]])
        out_exp = np.array(
            [
                [np.nan, 12.0, 6.0, 27.0],
                [np.nan, np.nan, 141.0, np.nan],
                [np.nan, 354.0, 135.0, 339.0],
            ]
        )
        np.testing.assert_array_equal(out_arr, out_exp)


class TestPrecipAccumulation:
    nc_pr = Path("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")
    nc_tasmin = Path("NRCANdaily", "nrcan_canada_daily_tasmin_1990.nc")

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_pr).pr  # mm/s
        prMM = open_dataset(self.nc_pr).pr
        prMM *= 86400
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.precip_accumulation(pr, freq="MS")
        out2 = atmos.precip_accumulation(prMM, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.precip_accumulation(pr, freq="MS")

        np.testing.assert_allclose(out1, out2, rtol=1e-7, atol=1e-4)
        np.testing.assert_allclose(out1, out3, rtol=1e-7, atol=1e-4)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0].values

        pr_tot = x1.sum()

        np.testing.assert_almost_equal(pr_tot, out1.values[0, 0, 0], 4)

        assert np.isnan(out1.values[0, 1, 0])
        assert np.isnan(out1.values[0, -1, -1])

    def test_with_different_phases(self, open_dataset):
        # test with different phases
        pr = open_dataset(self.nc_pr).pr  # mm/s
        tasmin = open_dataset(self.nc_tasmin).tasmin  # K

        out_tot = atmos.precip_accumulation(pr, freq="MS")
        out_sol = atmos.solid_precip_accumulation(pr, tas=tasmin, freq="MS")
        out_liq = atmos.liquid_precip_accumulation(pr, tas=tasmin, freq="MS")

        np.testing.assert_array_almost_equal(out_liq + out_sol, out_tot, 4)

        assert "solid" in out_sol.description
        assert "liquid" in out_liq.description
        assert out_sol.standard_name == "lwe_thickness_of_snowfall_amount"

        # With a non-default threshold
        out_sol = atmos.solid_precip_accumulation(
            pr, tas=tasmin, thresh="40 degF", freq="MS"
        )
        out_liq = atmos.liquid_precip_accumulation(
            pr, tas=tasmin, thresh="40 degF", freq="MS"
        )

        np.testing.assert_array_almost_equal(out_liq + out_sol, out_tot, 4)


class TestPrecipAverage:
    nc_pr = Path("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")
    nc_tasmin = Path("NRCANdaily", "nrcan_canada_daily_tasmin_1990.nc")

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_pr).pr  # mm/s
        prMM = open_dataset(self.nc_pr).pr
        prMM *= 86400
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.precip_average(pr, freq="MS")
        out2 = atmos.precip_average(prMM, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.precip_average(pr, freq="MS")

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 5)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0].values

        pr_mean = x1.mean()

        np.testing.assert_almost_equal(pr_mean, out1.values[0, 0, 0], 4)

        assert np.isnan(out1.values[0, 1, 0])

        assert np.isnan(out1.values[0, -1, -1])

    def test_with_different_phases(self, open_dataset):
        # test with different phases
        pr = open_dataset(self.nc_pr).pr  # mm/s
        tasmin = open_dataset(self.nc_tasmin).tasmin  # K

        out_tot = atmos.precip_average(pr, freq="MS")
        out_sol = atmos.solid_precip_average(pr, tas=tasmin, freq="MS")
        out_liq = atmos.liquid_precip_average(pr, tas=tasmin, freq="MS")

        np.testing.assert_array_almost_equal(out_liq + out_sol, out_tot, 4)

        assert "solid" in out_sol.description
        assert "liquid" in out_liq.description
        assert out_sol.standard_name == "lwe_average_of_snowfall_amount"

        # With a non-default threshold
        out_sol = atmos.solid_precip_average(
            pr, tas=tasmin, thresh="40 degF", freq="MS"
        )
        out_liq = atmos.liquid_precip_average(
            pr, tas=tasmin, thresh="40 degF", freq="MS"
        )

        np.testing.assert_array_almost_equal(out_liq + out_sol, out_tot, 4)


class TestStandardizedPrecip:
    nc_ds = Path("sdba", "CanESM2_1950-2100.nc")

    @pytest.mark.slow
    def test_3d_data_with_nans(self, open_dataset):
        # test with data
        ds = open_dataset(self.nc_ds)
        pr = ds.pr.sel(time=slice("2000"))  # kg m-2 s-1
        prMM = convert_units_to(pr, "mm/day", context="hydro")  # noqa
        # put a nan somewhere
        prMM.values[10] = np.nan
        pr.values[10] = np.nan

        out1 = atmos.standardized_precipitation_index(
            pr, freq="MS", window=1, dist="gamma", method="APP", fitkwargs={"floc": 0}
        )
        out2 = atmos.standardized_precipitation_index(
            prMM, freq="MS", window=1, dist="gamma", method="APP", fitkwargs={"floc": 0}
        )
        np.testing.assert_array_almost_equal(out1, out2, 3)

        # preparing water_budget for SPEI test
        with xr.set_options(keep_attrs=True):
            tasmax = ds.tasmax
            tas = tasmax - 2.5
            tasmin = tasmax - 5
            wb = xci.water_budget(pr, None, tasmin, tasmax, tas, None)
            wbMM = convert_units_to(wb, "mm/day", context="hydro")  # noqa

        out3 = atmos.standardized_precipitation_evapotranspiration_index(
            wb,
            freq="MS",
            window=1,
            dist="gamma",
            method="APP",
            fitkwargs={"floc": 0},
        )
        out4 = atmos.standardized_precipitation_evapotranspiration_index(
            wbMM,
            freq="MS",
            window=1,
            dist="gamma",
            method="APP",
            fitkwargs={"floc": 0},
        )

        np.testing.assert_array_almost_equal(out3, out4, 3)


class TestWetDays:
    nc_file = Path("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = "5 mm/d"
        out1 = atmos.wetdays(pr, thresh=pr_min, freq="MS")
        out2 = atmos.wetdays(prMM, thresh=pr_min, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.wetdays(pr, thresh=pr_min, freq="MS")

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0].values

        wd1 = (x1 >= int(pr_min.split(" ")[0])).sum()

        assert wd1 == out1.values[0, 0, 0]

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(wds.values[0, -1, -1]))


class TestWetPrcptot:
    """Testing of prcptot with wet days"""

    def test_simple(self, atmosds):
        pr = atmosds.pr

        thresh = "1 mm/day"
        out = atmos.wet_precip_accumulation(pr, thresh=thresh)

        # Reference value
        t = core.units.convert_units_to(thresh, pr, context="hydro")
        pa = atmos.precip_accumulation(pr.where(pr >= t, 0))
        np.testing.assert_array_equal(out, pa)


class TestDailyIntensity:
    # testing of wet_day and daily_pr_intensity, both are related

    nc_file = Path("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        # compute with both skipna options
        pr_min = "2 mm/d"
        # dis = daily_pr_intensity(pr, pr_min=pr_min, freq='MS', skipna=True)

        out1 = atmos.daily_pr_intensity(pr, thresh=pr_min, freq="MS")
        out2 = atmos.daily_pr_intensity(prMM, thresh=pr_min, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.daily_pr_intensity(pr, thresh=pr_min, freq="MS")

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 3)

        x1 = prMM[:31, 0, 0].values

        di1 = x1[x1 >= int(pr_min.split(" ")[0])].mean()
        # buffer = np.ma.masked_invalid(x2)
        # di2 = buffer[buffer >= pr_min].mean()

        assert np.allclose(di1, out1.values[0, 0, 0])
        # assert (np.allclose(di1, dis.values[0, 0, 0]))
        assert np.isnan(out1.values[0, 1, 0])
        # assert (np.allclose(di2, dis.values[0, 1, 0]))
        assert np.isnan(out1.values[0, -1, -1])
        # assert (np.isnan(dis.values[0, -1, -1]))


class TestMaxPrIntensity:
    def test_simple(self, pr_hr_series):
        pr1 = pr_hr_series(np.zeros(366 * 24))
        pr1[10:20] += np.arange(10)
        pr2 = pr_hr_series(np.ones(366 * 24))

        pr = xr.concat([pr1, pr2], dim="site")
        out = atmos.max_pr_intensity(pr, window=2, freq="YS")
        np.testing.assert_array_almost_equal(out.isel(time=0), [8.5 * 3600, 3600])


class TestMax1Day:
    # testing of wet_day and daily_pr_intensity, both are related

    nc_file = Path("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan

        out1 = atmos.max_1day_precipitation_amount(pr, freq="MS")
        out2 = atmos.max_1day_precipitation_amount(prMM, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.max_1day_precipitation_amount(pr, freq="MS")

        np.testing.assert_array_almost_equal(out1, out2, 3)
        np.testing.assert_array_almost_equal(out1, out3, 3)

        x1 = prMM[:31, 0, 0].values
        rx1 = x1.max()

        assert np.allclose(rx1, out1.values[0, 0, 0])
        assert np.isnan(out1.values[0, 1, 0])
        assert np.isnan(out1.values[0, -1, -1])


class TestMaxNDay:
    # testing of wet_day and daily_pr_intensity, both are related

    nc_file = Path("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")

    @pytest.mark.parametrize(
        "units,factor,chunks",
        [
            ("mm/day", 86400.0, None),
            ("kg m-2 s-1", 1, None),
            ("mm/s", 1, {"time": 73.0}),
        ],
    )
    def test_3d_data_with_nans(self, open_dataset, units, factor, chunks):
        # test with 3d data
        pr1 = open_dataset(self.nc_file).pr
        pr2 = open_dataset(self.nc_file, chunks=chunks).pr
        pr2.values *= factor
        pr2.attrs["units"] = units
        # put a nan somewhere
        pr2.values[10, 1, 0] = np.nan
        pr1.values[10, 1, 0] = np.nan
        wind = 3
        out1 = atmos.max_n_day_precipitation_amount(pr1, window=wind, freq="MS")
        out2 = atmos.max_n_day_precipitation_amount(pr2, window=wind, freq="MS")

        np.testing.assert_array_almost_equal(out1, out2, 3)

        x1 = pr1[:31, 0, 0].values * 86400
        df = pd.DataFrame({"pr": x1})
        rx3 = df.rolling(wind).sum().max()

        assert np.allclose(rx3, out1.values[0, 0, 0])
        assert np.isnan(out1.values[0, 1, 0])
        assert np.isnan(out1.values[0, -1, -1])


class TestMaxConsecWetDays:
    nc_file = Path("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = "5 mm/d"
        out1 = atmos.maximum_consecutive_wet_days(pr, thresh=pr_min, freq="MS")
        out2 = atmos.maximum_consecutive_wet_days(prMM, thresh=pr_min, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.maximum_consecutive_wet_days(pr, thresh=pr_min, freq="MS")

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0] * 0.0
        x1[5:10] = 10
        x1.attrs["units"] = "mm/day"
        cwd1 = atmos.maximum_consecutive_wet_days(x1, freq="MS")

        assert cwd1 == 5

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])


class TestMaxConsecDryDays:
    nc_file = Path("NRCANdaily", "nrcan_canada_daily_pr_1990.nc")

    def test_3d_data_with_nans(self, open_dataset):
        # test with 3d data
        pr = open_dataset(self.nc_file).pr
        prMM = open_dataset(self.nc_file).pr
        prMM.values *= 86400.0
        prMM.attrs["units"] = "mm/day"
        # put a nan somewhere
        prMM.values[10, 1, 0] = np.nan
        pr.values[10, 1, 0] = np.nan
        pr_min = "5 mm/d"
        out1 = atmos.maximum_consecutive_dry_days(pr, thresh=pr_min, freq="MS")
        out2 = atmos.maximum_consecutive_dry_days(prMM, thresh=pr_min, freq="MS")

        # test kg m-2 s-1
        pr.attrs["units"] = "kg m-2 s-1"
        out3 = atmos.maximum_consecutive_dry_days(pr, thresh=pr_min, freq="MS")

        np.testing.assert_array_equal(out1, out2)
        np.testing.assert_array_equal(out1, out3)

        # check some vector with and without a nan
        x1 = prMM[:31, 0, 0] * 0.0 + 50.0
        x1[5:10] = 0
        x1.attrs["units"] = "mm/day"
        cdd1 = atmos.maximum_consecutive_dry_days(x1, freq="MS")

        assert cdd1 == 5

        assert np.isnan(out1.values[0, 1, 0])

        # make sure that vector with all nans gives nans whatever skipna
        assert np.isnan(out1.values[0, -1, -1])


class TestSnowfallDate:
    tasmin_file = "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc"
    pr_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    @classmethod
    def get_snowfall(cls, open_dataset):
        dnr = xr.merge((open_dataset(cls.pr_file), open_dataset(cls.tasmin_file)))
        return atmos.snowfall_approximation(
            dnr.pr, tas=dnr.tasmin, thresh="-0.5 degC", method="binary"
        )

    def test_first_snowfall(self, open_dataset):
        with set_options(check_missing="skip"):
            fs = atmos.first_snowfall(
                prsn=self.get_snowfall(open_dataset), thresh="0.5 mm/day"
            )

        np.testing.assert_array_equal(
            fs[:, [0, 45, 82], [10, 105, 155]],
            np.array(
                [
                    [[1, 1, 1], [1, 1, 1], [11, np.nan, np.nan]],
                    [[254, 256, 277], [274, 292, 275], [300, np.nan, np.nan]],
                ]
            ),
        )

    def test_last_snowfall(self, open_dataset):
        with set_options(check_missing="skip"):
            ls = atmos.last_snowfall(
                prsn=self.get_snowfall(open_dataset), thresh="0.5 mm/day"
            )

        np.testing.assert_array_equal(
            ls[:, [0, 45, 82], [10, 105, 155]],
            np.array(
                [
                    [[155, 151, 129], [127, 157, 110], [106, np.nan, np.nan]],
                    [[365, 363, 363], [365.0, 364, 364], [362, np.nan, np.nan]],
                ]
            ),
        )


class TestDaysWithSnow:
    def test_simple(self, open_dataset, prsn_series):
        prsn = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").prsn
        out = atmos.days_with_snow(prsn, low="0 kg m-2 s-1", high="1e6 kg m-2 s-1")
        np.testing.assert_array_equal(out[1], [np.nan, 162, 159, 126, np.nan])


def test_days_over_precip_doy_thresh(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = percentile_doy(pr, window=5, per=80)

    out1 = atmos.days_over_precip_doy_thresh(pr, per)
    np.testing.assert_array_equal(out1[1, :, 0], np.array([81, 60, 68, 78]))

    out2 = atmos.days_over_precip_doy_thresh(pr, per, thresh="2 mm/d")
    np.testing.assert_array_equal(out2[1, :, 0], np.array([80, 59, 66, 78]))

    assert "only days with at least 2 mm/d are counted." in out2.description
    assert "[80]th percentile" in out2.attrs["description"]
    assert "['1990-01-01', '1993-12-31'] period" in out2.attrs["description"]
    assert "5 day(s)" in out2.attrs["description"]


def test_days_over_precip_thresh(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = pr.quantile(0.8, "time", keep_attrs=True)
    per.attrs["climatology_bounds"] = build_climatology_bounds(pr)

    out = atmos.days_over_precip_thresh(pr, per)

    np.testing.assert_allclose(
        out[1, :], np.array([80.0, 64.0, 65.0, 83.0]), atol=0.001
    )
    assert "80.0th percentile" in out.attrs["description"]
    assert "['1990-01-01', '1993-12-31'] period" in out.attrs["description"]


def test_days_over_precip_thresh__seasonal_indexer(open_dataset):
    # GIVEN
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = pr.quantile(0.8, "time", keep_attrs=True)
    # WHEN
    out = atmos.days_over_precip_thresh(
        pr, per, freq="YS", date_bounds=("01-10", "12-31")
    )
    # THEN
    np.testing.assert_almost_equal(out[0], np.array([81.0, 66.0, 66.0, 75.0]))


def test_fraction_over_precip_doy_thresh(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = percentile_doy(pr, window=5, per=80)

    out = atmos.fraction_over_precip_doy_thresh(pr, per)
    np.testing.assert_allclose(
        out[1, :, 0], np.array([0.803, 0.747, 0.745, 0.806]), atol=0.001
    )

    out = atmos.fraction_over_precip_doy_thresh(pr, per, thresh="0.002 m/d")
    np.testing.assert_allclose(
        out[1, :, 0], np.array([0.822, 0.780, 0.771, 0.829]), atol=0.001
    )

    assert "only days with at least 0.002 m/d are included" in out.description
    assert "[80]th percentile" in out.attrs["description"]
    assert "['1990-01-01', '1993-12-31'] period" in out.attrs["description"]
    assert "5 day(s)" in out.attrs["description"]


def test_fraction_over_precip_thresh(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    per = pr.quantile(0.8, "time", keep_attrs=True)
    per.attrs["climatology_bounds"] = build_climatology_bounds(pr)

    out = atmos.fraction_over_precip_thresh(pr, per)

    np.testing.assert_allclose(
        out[1, :], np.array([0.839, 0.812, 0.776, 0.864]), atol=0.001
    )

    assert "80.0th percentile" in out.attrs["description"]
    assert "['1990-01-01', '1993-12-31'] period" in out.attrs["description"]


def test_liquid_precip_ratio(open_dataset):
    ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")

    out = atmos.liquid_precip_ratio(pr=ds.pr, tas=ds.tas, thresh="0 degC", freq="YS")
    np.testing.assert_allclose(
        out[:, 0], np.array([0.919, 0.805, 0.525, 0.740, 0.993]), atol=1e3
    )

    with set_options(cf_compliance="raise"):
        # Test if tasmax is allowed
        out = atmos.liquid_precip_ratio(
            pr=ds.pr, tas=ds.tasmax, thresh="33 degF", freq="YS"
        )
        np.testing.assert_allclose(
            out[:, 0], np.array([0.975, 0.921, 0.547, 0.794, 0.999]), atol=1e3
        )
        assert "where temperature is above 33 degf." in out.description


def test_dry_spell(atmosds):
    pr = atmosds.pr

    events = atmos.dry_spell_frequency(pr, thresh="3 mm", window=7, freq="YS")
    total_d_sum = atmos.dry_spell_total_length(
        pr, thresh="3 mm", window=7, op="sum", freq="YS"
    )
    total_d_max = atmos.dry_spell_total_length(
        pr, thresh="3 mm", window=7, op="max", freq="YS"
    )
    max_d_sum = atmos.dry_spell_max_length(
        pr, thresh="3 mm", window=7, op="sum", freq="YS"
    )
    max_d_max = atmos.dry_spell_max_length(
        pr, thresh="3 mm", window=7, op="max", freq="YS"
    )
    total_d_sum = total_d_sum.sel(location="Halifax", drop=True).isel(time=slice(0, 2))
    total_d_max = total_d_max.sel(location="Halifax", drop=True).isel(time=slice(0, 2))
    max_d_sum = max_d_sum.sel(location="Halifax", drop=True).isel(time=slice(0, 2))
    max_d_max = max_d_max.sel(location="Halifax", drop=True).isel(time=slice(0, 2))

    np.testing.assert_allclose(events[0:2, 0], [5, 7], rtol=1e-1)
    np.testing.assert_allclose(total_d_sum, [50, 53], rtol=1e-1)
    np.testing.assert_allclose(total_d_max, [68, 97], rtol=1e-1)
    np.testing.assert_allclose(max_d_sum, [14, 10], rtol=1e-1)
    np.testing.assert_allclose(max_d_max, [14, 14], rtol=1e-1)

    assert (
        "The annual number of dry periods of 7 day(s) or more, "
        "during which the total precipitation on a window of 7 day(s) is below 3 mm."
    ) in events.description
    assert (
        "The annual number of days in dry periods of 7 day(s) or more"
        in total_d_sum.description
    )
    assert (
        "The annual number of days in dry periods of 7 day(s) or more"
        in total_d_max.description
    )
    assert (
        "The maximum annual number of consecutive days in a dry period of 7 day(s) or more"
        in max_d_sum.description
    )
    assert (
        "The maximum annual number of consecutive days in a dry period of 7 day(s) or more"
        in max_d_max.description
    )


def test_dry_spell_total_length_indexer(pr_series):
    pr = pr_series(
        [np.nan] + [1] * 4 + [0] * 10 + [1] * 350, start="1900-01-01", units="mm/d"
    )
    out = atmos.dry_spell_total_length(
        pr,
        window=7,
        op="sum",
        thresh="3 mm",
        freq="MS",
    )
    np.testing.assert_allclose(out, [np.nan] + [0] * 11)

    out = atmos.dry_spell_total_length(
        pr, window=7, op="sum", thresh="3 mm", freq="MS", date_bounds=("01-10", "12-31")
    )
    np.testing.assert_allclose(out, [9] + [0] * 11)


def test_dry_spell_max_length_indexer(pr_series):
    pr = pr_series(
        [np.nan] + [1] * 4 + [0] * 10 + [1] * 350, start="1900-01-01", units="mm/d"
    )
    out = atmos.dry_spell_max_length(
        pr,
        window=7,
        op="sum",
        thresh="3 mm",
        freq="MS",
    )
    np.testing.assert_allclose(out, [np.nan] + [0] * 11)

    out = atmos.dry_spell_total_length(
        pr, window=7, op="sum", thresh="3 mm", freq="MS", date_bounds=("01-10", "12-31")
    )
    np.testing.assert_allclose(out, [9] + [0] * 11)


def test_dry_spell_frequency_op(open_dataset):
    pr = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").pr
    test_sum = atmos.dry_spell_frequency(
        pr, thresh="3 mm", window=7, freq="MS", op="sum"
    )
    test_max = atmos.dry_spell_frequency(
        pr, thresh="3 mm", window=7, freq="MS", op="max"
    )

    np.testing.assert_allclose(
        test_sum[0, :14], [1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0], rtol=1e-1
    )
    np.testing.assert_allclose(
        test_max[0, :14], [1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 2, 1], rtol=1e-1
    )
    assert (
        "The monthly number of dry periods of 7 day(s) or more, "
        "during which the total precipitation on a window of 7 day(s) is below 3 mm."
    ) in test_sum.description
    assert (
        "The monthly number of dry periods of 7 day(s) or more, "
        "during which the maximal precipitation on a window of 7 day(s) is below 3 mm."
    ) in test_max.description


class TestSnowfallMeteoSwiss:
    tasmin_file = "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc"
    pr_file = "NRCANdaily/nrcan_canada_daily_pr_1990.nc"

    @classmethod
    def get_snowfall(cls, open_dataset):
        dnr = xr.merge((open_dataset(cls.pr_file), open_dataset(cls.tasmin_file)))
        return atmos.snowfall_approximation(
            dnr.pr, tas=dnr.tasmin, thresh="-0.5 degC", method="binary"
        )

    def test_snowfall_frequency(self, open_dataset):
        prsn = self.get_snowfall(open_dataset)
        with set_options(check_missing="skip"):
            sf = atmos.snowfall_frequency(prsn=prsn, thresh="1 mm/day")
        expected = np.array(
            [
                [
                    [27.624, 29.834, 25.414],
                    [22.652, 25.414, 22.652],
                    [12.155, 0.0, 0.0],
                ],
                [
                    [23.913, 23.370, 20.652],
                    [17.391, 15.761, 13.043],
                    [4.891, 0.0, 0.0],
                ],
            ]
        )
        np.testing.assert_allclose(
            sf[:, [0, 45, 82], [10, 105, 155]],
            expected,
            rtol=1e-3,
        )

    def test_snowfall_intensity(self, open_dataset):
        prsn = self.get_snowfall(open_dataset)
        with set_options(check_missing="skip"):
            si = atmos.snowfall_intensity(prsn=prsn, thresh="1 mm/day")
        expected = np.array(
            [
                [
                    [3.585, 3.839, 4.446],
                    [5.148, 5.884, 5.764],
                    [5.910, 0.0, 0.0],
                ],
                [
                    [5.093, 5.284, 5.539],
                    [5.946, 6.840, 9.702],
                    [9.522, 0.0, 0.0],
                ],
            ]
        )
        np.testing.assert_allclose(
            si[:, [0, 45, 82], [10, 105, 155]],
            expected,
            rtol=1e-3,
        )



"""Synoptic indicator definitions."""

from __future__ import annotations

from xclim import indices
from xclim.core.indicator import Indicator

__all__ = ["jetstream_metric_woollings"]


class JetStream(Indicator):
    """Indicator involving daily u- and/or v-component wind series."""

    src_freq = "D"


jetstream_metric_woollings = JetStream(
    title="Strength and latitude of jetstream",
    identifier="jetstream_metric_woollings",
    var_name=["jetlat", "jetstr"],
    units=["degrees_north", "m s-1"],
    long_name=[
        "Latitude of maximum smoothed zonal wind speed",
        "Maximum strength of smoothed zonal wind speed",
    ],
    description=[
        "Daily latitude of maximum Lanczos smoothed zonal wind speed.",
        "Daily maximum strength of Lanczos smoothed zonal wind speed.",
    ],
    compute=indices.jetstream_metric_woollings,
)



"""
Ensemble Robustness Metrics
===========================

Robustness metrics are used to estimate the confidence of the climate change signal of an ensemble.
This submodule is inspired by and tries to follow the guidelines of the IPCC,
more specifically :cite:p:`collins_long-term_2013` (AR5) and :cite:cts:`ipccatlas_ar6wg1` (AR6).
"""

from __future__ import annotations

from inspect import Parameter, signature

import numpy as np
import scipy.stats as spstats  # noqa
import xarray as xr

from xclim.core.formatting import gen_call_string, update_xclim_history
from xclim.indices.generic import compare, detrend

__all__ = [
    "robustness_categories",
    "robustness_coefficient",
    "robustness_fractions",
]


SIGNIFICANCE_TESTS = {}
"""Registry of change significance tests.

New tests must be decorated with :py:func:`significance_test` and fulfill the following requirements:

- Function name should begin by "_", registered test name is the function name without its first character and with _ replaced by -.
- Function must accept 2 positional arguments : fut and ref (see :py:func:`robustness_fractions` for definitions)
- Function may accept other keyword-only arguments.
- Function must return 2 values :
    + `changed` : 1D boolean array along `realization`. True for realization with significant change.
    + `pvals` : 1D float array along `realization`. P-values of the statistical test. Should be `None` for test where is doesn't apply.
"""


def significance_test(func):
    """Register a significance test for use in :py:func:`robustness_fractions`.

    See :py:data:`SIGNIFICANCE_TESTS`.
    """
    SIGNIFICANCE_TESTS[func.__name__[1:].replace("_", "-")] = func
    return func


# This function's docstring is modified to include the registered test names and docs.
# See end of this file.
@update_xclim_history
def robustness_fractions(  # noqa: C901
    fut: xr.DataArray,
    ref: xr.DataArray | None = None,
    test: str | None = None,
    weights: xr.DataArray | None = None,
    **kwargs,
) -> xr.Dataset:
    r"""Robustness statistics qualifying how members of an ensemble agree on the existence of change and on its sign.

    Parameters
    ----------
    fut : xr.DataArray
        Future period values along 'realization' and 'time' (..., nr, nt1)
        or if `ref` is None, Delta values along `realization` (..., nr).
    ref : xr.DataArray, optional
        Reference period values along realization' and 'time'  (..., nr, nt2).
        The size of the 'time' axis does not need to match the one of `fut`.
        But their 'realization' axes must be identical and the other coordinates should be the same.
        If `None` (default), values of `fut` are assumed to be deltas instead of
        a distribution across the future period.
    test : {tests_list}, optional
        Name of the statistical test used to determine if there was significant change. See notes.
    weights : xr.DataArray
        Weights to apply along the 'realization' dimension. This array cannot contain missing values.
    \*\*kwargs
        Other arguments specific to the statistical test. See notes.

    Returns
    -------
    xr.Dataset
        Same coordinates as `fut` and  `ref`, but no `time` and no `realization`.

        Variables:

        changed :
                The weighted fraction of valid members showing significant change.
                Passing `test=None` yields change_frac = 1 everywhere. Same type as `fut`.
        positive :
                The weighted fraction of valid members showing strictly positive change, no matter if it is significant or not.
        changed_positive :
                The weighted fraction of valid members showing significant and positive change.
        negative :
                The weighted fraction of valid members showing strictly negative change, no matter if it is significant or not.
        changed_negative :
                The weighted fraction of valid members showing significant and negative change.
        agree :
                The weighted fraction of valid members agreeing on the sign of change. It is the maximum between positive, negative and the rest.
        valid :
                The weighted fraction of valid members. A member is valid is there are no NaNs along the time axes of `fut` and  `ref`.
        pvals :
                The p-values estimated by the significance tests. Only returned if the test uses `pvals`. Has the  `realization` dimension.

    Notes
    -----
    The table below shows the coefficient needed to retrieve the number of members
    that have the indicated characteristics, by multiplying it by the total
    number of members (`fut.realization.size`) and by `valid_frac`, assuming uniform weights.
    For compactness, we rename the outputs cf, pf, cpf, nf and cnf.

    +-----------------+--------------------+------------------------+------------+
    |                 | Significant change | Non-significant change | Any change |
    +-----------------+--------------------+------------------------+------------+
    | Any direction   | cf                 | 1 - cf                 | 1          |
    +-----------------+--------------------+------------------------+------------+
    | Positive change | cpf                | pf - cpf               | pf         |
    +-----------------+--------------------+------------------------+------------+
    | Negative change | cnf                | nf - cnf               | nf         |
    +-----------------+--------------------+------------------------+------------+

    And members showing absolutely no change are ``1 - nf - pf``.

    Available statistical tests are :

    {tests_doc}
        threshold :
                Change is considered significant when it exceeds an absolute or relative threshold.
                Accepts one argument, either "abs_thresh" or "rel_thresh".
        None :
                Significant change is not tested. Members showing any positive change are
                included in the `pos_frac` output.

    References
    ----------
    :cite:cts:`tebaldi_mapping_2011`
    :cite:cts:`ipccatlas_ar6wg1`

    Example
    -------
    This example computes the mean temperature in an ensemble and compares two time
    periods, qualifying significant change through a single sample T-test.

    >>> from xclim import ensembles
    >>> ens = ensembles.create_ensemble(temperature_datasets)
    >>> tgmean = xclim.atmos.tg_mean(tas=ens.tas, freq="YS")
    >>> fut = tgmean.sel(time=slice("2020", "2050"))
    >>> ref = tgmean.sel(time=slice("1990", "2020"))
    >>> fractions = ensembles.robustness_fractions(fut, ref, test="ttest")
    """
    # Realization dimension name
    realization = "realization"

    # Assign dummy realization dimension if not present.
    if realization not in fut.dims:
        fut = fut.assign_coords({realization: "dummy"})
        fut = fut.expand_dims(realization)
    if ref is not None and realization not in ref.dims:
        ref = ref.assign_coords({realization: "dummy"})
        ref = ref.expand_dims(realization)

    # Get dummy weights to simplify code
    if weights is not None:
        w = weights
    else:
        w = xr.DataArray(
            [1] * fut[realization].size,
            dims=(realization,),
            coords={"realization": fut[realization]},
        )

    if ref is None:
        delta = fut
        valid = delta.notnull()
        if test not in [None, "threshold"]:
            raise ValueError(
                "When deltas are given (ref=None), 'test' must be None or 'threshold'."
            )
    else:
        delta = fut.mean("time") - ref.mean("time")
        valid = fut.notnull().all("time") & ref.notnull().all("time")

    if test is None:
        test_params = {}
        changed = xr.ones_like(delta).astype(bool)
        pvals = None
    elif test == "threshold":
        abs_thresh = kwargs.get("abs_thresh")
        rel_thresh = kwargs.get("rel_thresh")
        if abs_thresh is not None and rel_thresh is None:
            changed = abs(delta) > abs_thresh
            test_params = {"abs_thresh": abs_thresh}
        elif rel_thresh is not None and abs_thresh is None:
            changed = abs(delta / ref.mean("time")) > rel_thresh
            test_params = {"rel_thresh": rel_thresh}
        else:
            raise ValueError(
                "One and only one of abs_thresh or rel_thresh must be given if test='threshold'."
            )
        pvals = None
    elif test in SIGNIFICANCE_TESTS:
        test_func = SIGNIFICANCE_TESTS[test]
        test_params = {
            n: kwargs.get(n, p.default)
            for n, p in signature(test_func).parameters.items()
            if p.kind == Parameter.KEYWORD_ONLY
        }

        changed, pvals = test_func(fut, ref, **test_params)
    else:
        raise ValueError(
            f"Statistical test {test} must be one of {', '.join(SIGNIFICANCE_TESTS.keys())}."
        )

    valid_frac = valid.weighted(w).sum(realization) / fut[realization].size
    n_valid = valid.weighted(w).sum(realization)
    change_frac = changed.where(valid).weighted(w).sum(realization) / n_valid
    pos_frac = (delta > 0).where(valid).weighted(w).sum(realization) / n_valid
    neg_frac = (delta < 0).where(valid).weighted(w).sum(realization) / n_valid
    change_pos_frac = ((delta > 0) & changed).where(valid).weighted(w).sum(
        realization
    ) / n_valid
    change_neg_frac = ((delta < 0) & changed).where(valid).weighted(w).sum(
        realization
    ) / n_valid
    agree_frac = xr.concat((pos_frac, neg_frac, 1 - pos_frac - neg_frac), "sign").max(
        "sign"
    )

    # Metadata
    kwargs_str = gen_call_string("", **test_params)[1:-1]
    test_str = (
        f"Significant change was tested with test {test} and parameters {kwargs_str}."
    )

    out = xr.Dataset(
        {
            "changed": change_frac.assign_attrs(
                description="Fraction of members showing significant change. "
                + test_str,
                units="",
                test=str(test),
            ),
            "positive": pos_frac.assign_attrs(
                description="Fraction of valid members showing strictly positive change.",
                units="",
            ),
            "changed_positive": change_pos_frac.assign_attrs(
                description="Fraction of valid members showing significant and positive change. "
                + test_str,
                units="",
                test=str(test),
            ),
            "negative": neg_frac.assign_attrs(
                description="Fraction of valid members showing strictly negative change.",
                units="",
            ),
            "changed_negative": change_neg_frac.assign_attrs(
                description="Fraction of valid members showing significant and negative change. "
                + test_str,
                units="",
                test=str(test),
            ),
            "valid": valid_frac.assign_attrs(
                description="Fraction of valid members (No missing values along time).",
                units="",
            ),
            "agree": agree_frac.assign_attrs(
                description=(
                    "Fraction of valid members agreeing on the sign of change. "
                    "Maximum between the positive, negative and no change fractions."
                ),
                units="",
            ),
        },
        attrs={"description": "Significant change and sign of change fractions."},
    )

    if pvals is not None:
        pvals.attrs.update(
            description="P-values from change significance test. " + test_str,
            units="",
        )
        out = out.assign(pvals=pvals)

    # Keep attrs on non-modified coordinates
    for ncrd, crd in fut.coords.items():
        if ncrd in out.coords:
            out[ncrd].attrs.update(crd.attrs)

    return out


def robustness_categories(
    changed_or_fractions: xr.Dataset | xr.DataArray,
    agree: xr.DataArray | None = None,
    *,
    categories: list[str] | None = None,
    ops: list[tuple[str, str]] | None = None,
    thresholds: list[tuple[float, float]] | None = None,
) -> xr.DataArray:
    """Create a categorical robustness map for mapping hatching patterns.

    Each robustness category is defined by a double threshold, one on the fraction of members showing significant
    change (`change_frac`) and one on the fraction of member agreeing on the sign of change (`agree_frac`).
    When the two thresholds are fulfilled, the point is assigned to the given category.
    The default values for the comparisons are the ones suggested by the IPCC for its "Advanced approach" described
    in the Cross-Chapter Box 1 of the Atlas of the AR6 WGI report (:cite:t:`ipccatlas_ar6wg1`).

    Parameters
    ----------
    changed_or_fractions : xr.Dataset or xr.DataArray
        Either the fraction of members showing significant change as an array or
        directly the output of :py:func:`robustness_fractions`.
    agree : xr.DataArray, optional
        The fraction of members agreeing on the sign of change. Only needed if the first argument is
        the `changed` array.
    categories : list of str, optional
        The label of each robustness categories. They are stored in the semicolon separated flag_descriptions
        attribute as well as in a compressed form in the flag_meanings attribute.
        If a point is mapped to two categories, priority is given to the first one in this list.
    ops : list of tuples of str, optional
        For each category, the comparison operators for `change_frac` and `agree_frac`.
        None or an empty string means the variable is not needed for this category.
    thresholds : list of tuples of float, optional
        For each category, the threshold to be used with the corresponding operator. All should be between 0 and 1.

    Returns
    -------
    xr.DataArray
        Categorical (int) array following the flag variables CF conventions.
        99 is used as a fill value for points that do not fall in any category.
    """
    if categories is None:
        categories = [
            "Robust signal",
            "No change or no signal",
            "Conflicting signal",
        ]

    if ops is None:
        ops = [(">=", ">="), ("<", None), (">=", "<")]

    if thresholds is None:
        thresholds = [(0.66, 0.8), (0.66, None), (0.66, 0.8)]

    src = changed_or_fractions.copy()  # Ensure no inplace changing of coords...
    if isinstance(src, xr.Dataset):
        # Output of robustness fractions
        changed = src.changed
        agree = src.agree
    else:
        changed = src

    # Initial map is all 99, same shape as change_frac
    robustness = (changed.copy() * 0).astype(int) + 99
    # We go in reverse gear so that the first categories have precedence in the case of multiple matches.
    for i, ((chg_op, agr_op), (chg_thresh, agr_thresh)) in reversed(
        list(enumerate(zip(ops, thresholds, strict=False), 1))
    ):
        if not agr_op:
            cond = compare(changed, chg_op, chg_thresh)
        elif not chg_op:
            cond = compare(agree, agr_op, agr_thresh)
        else:
            cond = compare(changed, chg_op, chg_thresh) & compare(
                agree, agr_op, agr_thresh
            )
        robustness = xr.where(~cond, robustness, i, keep_attrs=True)

    robustness = robustness.assign_attrs(
        flag_values=list(range(1, len(categories) + 1)),
        _FillValue=99,
        flag_descriptions=categories,
        flag_meanings=" ".join(
            map(lambda cat: cat.casefold().replace(" ", "_"), categories)
        ),
    )
    return robustness


@update_xclim_history
def robustness_coefficient(
    fut: xr.DataArray | xr.Dataset, ref: xr.DataArray | xr.Dataset
) -> xr.DataArray | xr.Dataset:
    """Robustness coefficient quantifying the robustness of a climate change signal in an ensemble.

    Taken from :cite:ts:`knutti_robustness_2013`.

    The robustness metric is defined as R = 1 − A1 / A2 , where A1 is defined as the integral of the squared area
    between two cumulative density functions characterizing the individual model projections and the multimodel mean
    projection and A2 is the integral of the squared area between two cumulative density functions characterizing
    the multimodel mean projection and the historical climate.
    Description taken from :cite:t:`knutti_robustness_2013`.

    A value of R equal to one implies perfect model agreement. Higher model spread or
    smaller signal decreases the value of R.

    Parameters
    ----------
    fut : Union[xr.DataArray, xr.Dataset]
        Future ensemble values along 'realization' and 'time' (nr, nt). Can be a dataset,
        in which case the coefficient is computed on each variable.
    ref : Union[xr.DataArray, xr.Dataset]
        Reference period values along 'time' (nt). Same type as `fut`.

    Returns
    -------
    xr.DataArray or xr.Dataset
        The robustness coefficient, ]-inf, 1], float. Same type as `fut` or `ref`.

    References
    ----------
    :cite:cts:`knutti_robustness_2013`
    """

    def _knutti_sedlacek(reference, future):
        def diff_cdf_sq_area_int(x1, x2):
            """Exact integral of the squared area between the non-parametric CDFs of 2 vectors."""
            # Non-parametric CDF on points x1 and x2
            # i.e. y1(x) is the proportion of x1 <= x
            y1 = (np.arange(x1.size) + 1) / x1.size
            y2 = (np.arange(x2.size) + 1) / x2.size

            x2_in_1 = np.searchsorted(x1, x2, side="right")  # Where to insert x2 in x1
            x1_in_2 = np.searchsorted(x2, x1, side="right")  # Where to insert x1 in x2

            # Merge to get all "discontinuities" of the CDF difference
            # y1 with repeated value (to the right) where x2 is inserted
            # Same for y2. 0s are prepended where needed.
            x = np.insert(x1, x2_in_1, x2)
            y1_f = np.insert(y1, x2_in_1, np.r_[0, y1][x2_in_1])
            y2_f = np.insert(y2, x1_in_2, np.r_[0, y2][x1_in_2])

            # Discrete integral of the squared difference (distance) between the two CDFs.
            return np.sum(np.diff(x) * (y1_f - y2_f)[:-1] ** 2)

        # Get sorted vectors
        v_fut = np.sort(future.flatten())  # "cumulative" models distribution
        v_favg = np.sort(future.mean(axis=-1))  # Multimodel mean
        v_ref = np.sort(reference)  # Historical values

        A1 = diff_cdf_sq_area_int(v_fut, v_favg)  # noqa
        A2 = diff_cdf_sq_area_int(v_ref, v_favg)  # noqa

        return 1 - A1 / A2

    R = xr.apply_ufunc(  # noqa
        _knutti_sedlacek,
        ref,
        fut,
        input_core_dims=[["time"], ["realization", "time"]],
        exclude_dims={"time"},
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )
    R.attrs.update(
        name="R",
        long_name="Ensemble robustness coefficient",
        description="Ensemble robustness coefficient as defined by Knutti and Sedláček (2013).",
        reference="Knutti, R. and Sedláček, J. (2013) Robustness and uncertainties in the new CMIP5 climate model projections. Nat. Clim. Change.",
        units="",
    )
    return R


@significance_test
def _ttest(fut, ref, *, p_change=0.05):
    """Single sample T-test. Same test as used by :cite:t:`tebaldi_mapping_2011`.

    The future values are compared against the reference mean (over 'time').
    Accepts argument p_change (float, default : 0.05) the p-value threshold for rejecting the hypothesis of no significant change.
    """

    def _ttest_func(f, r):
        # scipy>=1.9: popmean.axis[-1] must equal 1 for both fut and ref
        if np.isnan(f).all() or np.isnan(r).all():
            return np.nan

        return spstats.ttest_1samp(f, r[..., np.newaxis], axis=-1, nan_policy="omit")[1]

    # Test hypothesis of no significant change
    pvals = xr.apply_ufunc(
        _ttest_func,
        fut,
        ref.mean("time"),
        input_core_dims=[["time"], []],
        output_core_dims=[[]],
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )
    # When p < p_change, the hypothesis of no significant change is rejected.
    changed = pvals < p_change
    return changed, pvals


@significance_test
def _welch_ttest(fut, ref, *, p_change=0.05):
    """Two-sided T-test, without assuming equal population variance.

    Same significance criterion and argument as 'ttest'.
    """

    # Test hypothesis of no significant change
    # equal_var=False -> Welch's T-test
    def wtt_wrapper(f, r):  # This specific test can't manage an all-NaN slice
        if np.isnan(f).all() or np.isnan(r).all():
            return np.nan
        return spstats.ttest_ind(f, r, axis=-1, equal_var=False, nan_policy="omit")[1]

    pvals = xr.apply_ufunc(
        wtt_wrapper,
        fut,
        ref,
        input_core_dims=[["time"], ["time"]],
        output_core_dims=[[]],
        exclude_dims={"time"},
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )

    # When p < p_change, the hypothesis of no significant change is rejected.
    changed = pvals < p_change
    return changed, pvals


@significance_test
def _mannwhitney_utest(ref, fut, *, p_change=0.05):
    """Two-sided Mann-Whiney U-test. Same significance criterion and argument as 'ttest'."""

    def mwu_wrapper(f, r):  # This specific test can't manage an all-NaN slice
        if np.isnan(f).all() or np.isnan(r).all():
            return np.nan
        return spstats.mannwhitneyu(f, r, axis=-1, nan_policy="omit")[1]

    pvals = xr.apply_ufunc(
        mwu_wrapper,
        fut,
        ref,
        input_core_dims=[["time"], ["time"]],
        output_core_dims=[[]],
        exclude_dims={"time"},
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )
    # When p < p_change, the hypothesis of no significant change is rejected.
    changed = pvals < p_change
    return changed, pvals


@significance_test
def _brownforsythe_test(fut, ref, *, p_change=0.05):
    """Brown-Forsythe test assuming skewed, non-normal distributions.

    Same significance criterion and argument as 'ttest'.
    """
    pvals = xr.apply_ufunc(
        lambda f, r: spstats.levene(f, r, center="median")[1],
        fut,
        ref,
        input_core_dims=[["time"], ["time"]],
        output_core_dims=[[]],
        exclude_dims={"time"},
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )
    # When p < p_change, the hypothesis of no significant change is rejected.
    changed = pvals < p_change
    return changed, pvals


@significance_test
def _ipcc_ar6_c(fut, ref, *, ref_pi=None):
    r"""The advanced approach used in the IPCC Atlas chapter (:cite:t:`ipccatlas_ar6wg1`).

    Change is considered significant if the delta exceeds a threshold related to the internal variability.
    If pre-industrial data is given in argument `ref_pi`, the threshold is defined as
    :math:`\sqrt{2}*1.645*\sigma_{20yr}`, where :math:`\sigma_{20yr}` is the standard deviation of 20-year
    means computed from non-overlapping periods after detrending with a quadratic fit.
    Otherwise, when such pre-industrial control data is not available, the threshold is defined in relation to
    the historical data (`ref`) as :math:`\sqrt{\frac{2}{20}}*1.645*\sigma_{1yr}, where :math:`\sigma_{1yr}`
    is the inter-annual standard deviation measured after linearly detrending the data.
    See notebook :ref:`notebooks/ensembles:Ensembles` for more details.
    """
    # Ensure annual
    refy = ref.resample(time="YS").mean()
    if ref_pi is None:
        ref_detrended = detrend(refy, dim="time", deg=1)
        gamma = np.sqrt(2 / 20) * 1.645 * ref_detrended.std("time")
    else:
        ref_detrended = detrend(refy, dim="time", deg=2)
        gamma = (
            np.sqrt(2) * 1.645 * ref_detrended.resample(time="20YS").mean().std("time")
        )

    delta = fut.mean("time") - ref.mean("time")
    changed = abs(delta) > gamma
    return changed, None


# Add doc of each significance test to `robustness_fractions` output's doc.
def _gen_test_entry(namefunc):
    name, func = namefunc
    doc = func.__doc__.replace("\n    ", "\n\t\t").rstrip()
    return f"\t{name}:\n\t\t{doc}"


robustness_fractions.__doc__ = robustness_fractions.__doc__.format(
    tests_list="{" + ", ".join(list(SIGNIFICANCE_TESTS.keys()) + ["threshold"]) + "}",
    tests_doc="\n".join(map(_gen_test_entry, SIGNIFICANCE_TESTS.items())),
)
